# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for ChangeState event."""

import collections
from typing import Callable

from launch.event import Event

import lifecycle_msgs.msg

if False:
    # imports here would cause loops, but are only used as forward-references for type-checking
    from ...actions import LifecycleNode  # noqa: F401


class ChangeState(Event):
    """Event emitted when a state transition is requested for a lifecycle node."""

    name = 'launch_ros.events.lifecycle.ChangeState'
    valid_transitions = collections.OrderedDict([
        (lifecycle_msgs.msg.Transition.TRANSITION_CREATE, 'TRANSITION_CREATE'),
        (lifecycle_msgs.msg.Transition.TRANSITION_CONFIGURE, 'TRANSITION_CONFIGURE'),
        (lifecycle_msgs.msg.Transition.TRANSITION_CLEANUP, 'TRANSITION_CLEANUP'),
        (lifecycle_msgs.msg.Transition.TRANSITION_ACTIVATE, 'TRANSITION_ACTIVATE'),
        (lifecycle_msgs.msg.Transition.TRANSITION_DEACTIVATE, 'TRANSITION_DEACTIVATE'),
        (
            lifecycle_msgs.msg.Transition.TRANSITION_UNCONFIGURED_SHUTDOWN,
            'TRANSITION_UNCONFIGURED_SHUTDOWN',
        ),
        (
            lifecycle_msgs.msg.Transition.TRANSITION_INACTIVE_SHUTDOWN,
            'TRANSITION_INACTIVE_SHUTDOWN',
        ),
        (lifecycle_msgs.msg.Transition.TRANSITION_ACTIVE_SHUTDOWN, 'TRANSITION_ACTIVE_SHUTDOWN'),
        (lifecycle_msgs.msg.Transition.TRANSITION_DESTROY, 'TRANSITION_DESTROY'),
    ])
    valid_states = collections.OrderedDict([
        (lifecycle_msgs.msg.State.PRIMARY_STATE_UNKNOWN, 'PRIMARY_STATE_UNKNOWN'),
        (lifecycle_msgs.msg.State.PRIMARY_STATE_UNCONFIGURED, 'PRIMARY_STATE_UNCONFIGURED'),
        (lifecycle_msgs.msg.State.PRIMARY_STATE_INACTIVE, 'PRIMARY_STATE_INACTIVE'),
        (lifecycle_msgs.msg.State.PRIMARY_STATE_ACTIVE, 'PRIMARY_STATE_ACTIVE'),
        (lifecycle_msgs.msg.State.PRIMARY_STATE_FINALIZED, 'PRIMARY_STATE_FINALIZED'),
        (lifecycle_msgs.msg.State.TRANSITION_STATE_CONFIGURING, 'TRANSITION_STATE_CONFIGURING'),
        (lifecycle_msgs.msg.State.TRANSITION_STATE_CLEANINGUP, 'TRANSITION_STATE_CLEANINGUP'),
        (lifecycle_msgs.msg.State.TRANSITION_STATE_SHUTTINGDOWN, 'TRANSITION_STATE_SHUTTINGDOWN'),
        (lifecycle_msgs.msg.State.TRANSITION_STATE_ACTIVATING, 'TRANSITION_STATE_ACTIVATING'),
        (lifecycle_msgs.msg.State.TRANSITION_STATE_DEACTIVATING, 'TRANSITION_STATE_DEACTIVATING'),
        (
            lifecycle_msgs.msg.State.TRANSITION_STATE_ERRORPROCESSING,
            'TRANSITION_STATE_ERRORPROCESSING',
        ),
    ])

    def __init__(
        self,
        *,
        lifecycle_node_matcher: Callable[['LifecycleNode'], bool],
        transition_id: int
    ) -> None:
        """
        Create a ChangeState event.

        :param: lifecycle_node_matcher is a callable which returns True if the
            given lifecycle node should be affected by this event.
        :param: transition_id is the id of the requested transition which are
            defined in the :class:`lifecycle_msgs.msg.Transition` message class,
            e.g. `lifecycle_msgs.msg.Transition.TRANSITION_CONFIGURE`.
        """
        super().__init__()
        self.__lifecycle_node_matcher = lifecycle_node_matcher
        self.__transition_id = transition_id
        if transition_id not in self.valid_transitions.keys():
            raise ValueError("given transition_id of '{}', expected one of {{{}}}".format(
                transition_id,
                ', '.join(['{}: {}'.format(v, k) for k, v in self.valid_transitions.items()]),
            ))

    @property
    def lifecycle_node_matcher(self) -> Callable[['LifecycleNode'], bool]:
        """Getter for lifecycle_node_matcher."""
        return self.__lifecycle_node_matcher

    @property
    def transition_id(self) -> int:
        """Getter for transition_id."""
        return self.__transition_id



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for StateTransition event."""

from typing import Text

from launch.event import Event

import lifecycle_msgs.msg

if False:
    # imports here would cause loops, but are only used as forward-references for type-checking
    from ...actions import LifecycleNode  # noqa: F401


class StateTransition(Event):
    """Event emitted when a lifecycle node makes a state transition."""

    name = 'launch_ros.events.lifecycle.StateTransition'

    def __init__(
        self,
        *,
        action: 'LifecycleNode',
        msg: lifecycle_msgs.msg.TransitionEvent
    ) -> None:
        """
        Create a StateTransition event.

        :param: action the instance of class::`LifecycleNode` that generated this event
        :param: msg the instance of the ROS message TransitionEvent that generated this event
        """
        super().__init__()
        self.__action = action
        self.__msg = msg
        self.__timestamp = msg.timestamp
        self.__transition = msg.transition.label
        self.__start_state = msg.start_state.label
        self.__goal_state = msg.goal_state.label

    @property
    def action(self) -> 'LifecycleNode':
        """Getter for action."""
        return self.__action

    @property
    def msg(self) -> lifecycle_msgs.msg.TransitionEvent:
        """Getter for msg."""
        return self.__msg

    @property
    def timestamp(self) -> int:
        """Getter for timestamp."""
        return self.__timestamp

    @property
    def transition(self) -> Text:
        """Getter for transition."""
        return self.__transition

    @property
    def start_state(self) -> Text:
        """Getter for start_state."""
        return self.__start_state

    @property
    def goal_state(self) -> Text:
        """Getter for goal_state."""
        return self.__goal_state



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the RosTimer action."""

import asyncio
import collections.abc
from functools import partial
from typing import Iterable
from typing import Optional
from typing import Text
from typing import Union

from launch.actions import TimerAction
from launch.events import TimerEvent
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.launch_description_entity import LaunchDescriptionEntity
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.some_substitutions_type import SomeSubstitutionsType_types_tuple
from launch.utilities import create_future
from launch.utilities import ensure_argument_type
from launch.utilities import type_utils

from launch_ros.ros_adapters import get_ros_node


@expose_action('ros_timer')
class RosTimer(TimerAction):
    """
    Action that defers other entities until a period of time has passed, unless canceled.

    This timer uses ROS time instead of wall clock time.
    To enable the use of sim time, you must also use the SetUseSimTime action.
    All timers are "one-shot", in that they only fire one time and never again.
    """

    def __init__(
        self,
        *,
        period: Union[float, SomeSubstitutionsType],
        actions: Iterable[LaunchDescriptionEntity],
        **kwargs
    ) -> None:
        """
        Create a RosTimer.

        :param period: is the time (in seconds) to set the timer for.
        :param actions: is an iterable containing actions to be executed upon on timeout.
        """
        super().__init__(period=period, actions=actions, **kwargs)
        period_types = list(SomeSubstitutionsType_types_tuple) + [float]
        ensure_argument_type(period, period_types, 'period', 'RosTimer')
        ensure_argument_type(actions, collections.abc.Iterable, 'actions', 'RosTimer')
        self.__period = type_utils.normalize_typed_substitution(period, float)
        self.__timer_future: Optional[asyncio.Future] = None

    def __timer_callback(self):
        if not self.__timer_future.done():
            self.__timer_future.set_result(True)

    async def _wait_to_fire_event(self, context):
        node = get_ros_node(context)
        node.create_timer(
            type_utils.perform_typed_substitution(context, self.__period, float),
            partial(context.asyncio_loop.call_soon_threadsafe, self.__timer_callback),
        )

        done, pending = await asyncio.wait(
            [self._canceled_future, self.__timer_future],
            return_when=asyncio.FIRST_COMPLETED
        )

        if not self._canceled_future.done():
            await context.emit_event(TimerEvent(timer_action=self))
        self._completed_future.set_result(None)

    @classmethod
    def parse(
        cls,
        entity: Entity,
        parser: Parser,
    ):
        """Return the `RosTimer` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['period'] = parser.parse_if_substitutions(
            entity.get_attr('period', data_type=float, can_be_str=True))
        kwargs['actions'] = [parser.parse_action(child) for child in entity.children]
        return cls, kwargs

    def describe(self) -> Text:
        """Return a description of this RosTimer."""
        return 'RosTimer(period={}, actions=<actions>)'.format(self.__period)

    def execute(self, context: LaunchContext):
        self.__timer_future = create_future(context.asyncio_loop)
        return super().execute(context)



# Copyright 2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with utility functions for handling namespaces."""


from typing import Optional
from typing import Text
from typing import Type
from typing import TypeVar


def is_root_namespace(ns: Text) -> bool:
    """Return `True` if `ns` is `'/'`."""
    return ns == '/'


def is_namespace_absolute(ns: Text) -> bool:
    """Return `True` if `ns` is absolute."""
    return ns.startswith('/')


def prefix_namespace(
    base_ns: Optional[Text],
    ns: Optional[Text],
) -> Optional[Text]:
    """
    Return `ns` prefixed with `base_ns` if `ns` is relative, return `ns` if not.

    :param `base_ns`: Prefix to be added to `ns`.
    :param `ns`: Namespace to be prefixed.
    :return: `None` if both `base_ns` and `ns` are `None`, or
        `base_ns` if `ns` is `None`, or
        `ns` if `base_ns` is `None`, or
        `ns` if `ns` is absolute, or
        `ns` prefixed with `base_ns`.
        In all cases, trailing `/` are stripped from the result.

    ## Examples:

    ```python3
    combined_ns = prefix_namespace('my_ns', 'original_ns')
    assert combined_ns == 'my_ns/original_ns'

    combined_ns = prefix_namespace('/my_ns', 'original_ns')
    assert combined_ns == '/my_ns/original_ns'

    combined_ns = prefix_namespace('my_ns', '/original_ns')
    assert combined_ns == '/original_ns'

    combined_ns = prefix_namespace(None, 'original_ns')
    assert combined_ns == 'original_ns'

    combined_ns = prefix_namespace('my_ns', None)
    assert combined_ns == 'my_ns'
    ```
    """
    combined_ns: Optional[Text] = None
    if base_ns is not None or ns is not None:
        if ns is None:
            combined_ns = base_ns
        elif not base_ns or is_namespace_absolute(ns):
            combined_ns = ns
        else:
            if is_root_namespace(base_ns):
                base_ns = ''
            combined_ns = (
                base_ns + '/' + ns
            )
        if not is_root_namespace(combined_ns):
            combined_ns = combined_ns.rstrip('/')
    return combined_ns


OptionalText = TypeVar('OptionalText', Text, Type[None])


def make_namespace_absolute(ns: OptionalText) -> OptionalText:
    """Make a relative namespace absolute."""
    if ns is None:
        return None
    if not is_namespace_absolute(ns):
        return '/' + ns
    return ns



# Copyright 2022 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the Parameter substitution."""

from typing import Iterable
from typing import List
from typing import Text

from launch.frontend import expose_substitution
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitution import Substitution
from launch.substitutions.substitution_failure import SubstitutionFailure
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions


@expose_substitution('param')
class Parameter(Substitution):
    """
    Substitution that tries to get a parameter that was set by SetParameter.

    :raise: SubstitutionFailure when param is not found
    """

    def __init__(
        self,
        name: SomeSubstitutionsType,
    ) -> None:
        """Create a Parameter substitution."""
        super().__init__()
        self.__name = normalize_to_list_of_substitutions(name)

    @classmethod
    def parse(cls, data: Iterable[SomeSubstitutionsType]):
        """Parse a Parameter substitution."""
        if not data or len(data) != 1:
            raise AttributeError('param substitutions expect 1 argument')
        kwargs = {'name': data[0]}
        return cls, kwargs

    @property
    def name(self) -> List[Substitution]:
        """Getter for name."""
        return self.__name

    def describe(self) -> Text:
        """Return a description of this substitution as a string."""
        name_str = ' + '.join([sub.describe() for sub in self.name])
        return '{}(name={})'.format(self.__class__.__name__, name_str)

    def perform(self, context: LaunchContext) -> Text:
        """Perform the substitution."""
        name = perform_substitutions(context, self.name)
        params_container = context.launch_configurations.get('global_params', None)
        for param in params_container:
            if isinstance(param, tuple):
                if param[0] == name:
                    return param[1]
        raise SubstitutionFailure("parameter '{}' not found".format(name))



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for normalize_remap_rule utility."""

from launch.utilities import ensure_argument_type
from launch.utilities import normalize_to_list_of_substitutions

from ..remap_rule_type import RemapRule
from ..remap_rule_type import RemapRules
from ..remap_rule_type import SomeRemapRule
from ..remap_rule_type import SomeRemapRules


def normalize_remap_rule(remap_rule: SomeRemapRule) -> RemapRule:
    """Normalize a remap rule to a specific type."""
    ensure_argument_type(remap_rule, (tuple), 'remap_rule')
    if len(remap_rule) != 2:
        raise TypeError(
            'remap_rule must be a tuple of length 2, got length {}'.format(len(remap_rule)))
    from_sub = tuple(normalize_to_list_of_substitutions(remap_rule[0]))
    to_sub = tuple(normalize_to_list_of_substitutions(remap_rule[1]))
    return from_sub, to_sub


def normalize_remap_rules(remap_rules: SomeRemapRules) -> RemapRules:
    """Normalize multiple remap rules."""
    yield from map(normalize_remap_rule, remap_rules)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with utility for normalizing parameters to a node."""

from collections.abc import Mapping
from collections.abc import Sequence
from typing import cast
from typing import List
from typing import Optional
from typing import Sequence as SequenceTypeHint
from typing import Set  # noqa: F401
from typing import Union

from launch.some_substitutions_type import SomeSubstitutionsType
from launch.some_substitutions_type import SomeSubstitutionsType_types_tuple
from launch.substitution import Substitution
from launch.substitutions import TextSubstitution
from launch.utilities import ensure_argument_type
from launch.utilities import normalize_to_list_of_substitutions

import yaml

from ..parameter_descriptions import Parameter as ParameterDescription
from ..parameter_descriptions import ParameterFile
from ..parameter_descriptions import ParameterValue as ParameterValueDescription
from ..parameters_type import Parameters
from ..parameters_type import ParametersDict
from ..parameters_type import ParameterValue
from ..parameters_type import SomeParameters
from ..parameters_type import SomeParametersDict
from ..parameters_type import SomeParameterValue


def _normalize_parameter_array_value(value: SomeParameterValue) -> ParameterValue:
    """Normalize substitutions while preserving the type if it's not a substitution."""
    if not isinstance(value, Sequence):
        raise TypeError('Value {} must be a sequence'.format(repr(value)))

    # Figure out what type the list should be
    has_types = set()  # type: Set[type]
    for subvalue in value:
        allowed_subtypes = (float, int, str, bool) + SomeSubstitutionsType_types_tuple
        ensure_argument_type(subvalue, allowed_subtypes, 'subvalue')

        if isinstance(subvalue, Substitution):
            has_types.add(Substitution)
        elif isinstance(subvalue, str):
            has_types.add(str)
        elif isinstance(subvalue, bool):
            has_types.add(bool)
        elif isinstance(subvalue, int):
            has_types.add(int)
        elif isinstance(subvalue, float):
            has_types.add(float)
        elif isinstance(subvalue, Sequence):
            has_types.add(tuple)
        else:
            raise RuntimeError('Failed to handle type {}'.format(repr(subvalue)))

    if {int} == has_types:
        # everything is an integer
        make_mypy_happy_int = cast(List[int], value)
        return tuple(int(e) for e in make_mypy_happy_int)
    elif has_types in ({float}, {int, float}):
        # all were floats or ints, so return floats
        make_mypy_happy_float = cast(List[Union[int, float]], value)
        return tuple(float(e) for e in make_mypy_happy_float)
    elif Substitution in has_types and has_types.issubset({str, Substitution}):
        # make a list of substitutions forming a single string
        return tuple(normalize_to_list_of_substitutions(cast(SomeSubstitutionsType, value)))
    elif {bool} == has_types:
        # all where bools
        return tuple(bool(e) for e in value)
    else:
        # Should evaluate to a list of strings
        # Normalize to a list of lists of substitutions
        new_value = []  # type: List[SomeSubstitutionsType]
        for element in value:
            if isinstance(element, (float, int, bool, str)):
                new_value.append(yaml.dump(element))
            else:
                new_value.append(element)
        return tuple(normalize_to_list_of_substitutions(e) for e in new_value)


def normalize_parameter_dict(
    parameters: SomeParametersDict, *,
    _prefix: Optional[SequenceTypeHint[Substitution]] = None
) -> ParametersDict:
    """
    Normalize types used to store parameters in a dictionary.

    The parameters are passed as a dictionary that specifies parameter rules.
    Keys of the dictionary can be strings, a Substitution, or an iterable of Substitution.
    A normalized key will be a tuple of substitutions.
    Values in the dictionary can be strings, integers, floats, substitutions, lists of
    the previous types, or another dictionary with the same properties.

    Normalized values that were lists will have all subvalues converted to the same type.
    If all subvalues are int or float, then the normalized subvalues will all be float.
    If the subvalues otherwise do not all have the same type, then the normalized subvalues
    will be lists of Substitution that will result in a list of strings.

    Values that are a list of strings will become a list of strings when normalized and evaluated.
    Values that are a list which has at least one :class:`Substitution` and all other elements
    are either strings or a list of substitutions will become a single list of substitutions that
    will evaluate to a single string.
    To make a list of strings from substitutions, each item in the list must be a list or tuple.

    Normalized values that contained nested dictionaries will be collapsed into a single
    layer with parameter names concatenated with the parameter namespace separator ".".

    :param parameters: Parameters to be normalized
    :param _prefix: internal parameter used for flatening recursive dictionaries
    :return: Normalized parameters
    """
    if not isinstance(parameters, Mapping):
        raise TypeError('expected dict')

    normalized = {}  # type: ParametersDict
    for name, value in parameters.items():
        # First make name a list of substitutions
        name = normalize_to_list_of_substitutions(name)
        if _prefix:
            # Prefix name if there is a recursive dictionary
            # weird looking logic to combine into one list to appease mypy
            combined = list(_prefix)
            combined.append(TextSubstitution(text='.'))
            combined.extend(name)
            name = combined

        # Normalize the value next
        if isinstance(value, Mapping):
            # Flatten recursive dictionaries
            sub_dict = normalize_parameter_dict(value, _prefix=name)
            normalized.update(sub_dict)
        elif isinstance(value, ParameterValueDescription):
            normalized[tuple(name)] = value
        elif isinstance(value, str):
            normalized[tuple(name)] = tuple(normalize_to_list_of_substitutions(yaml.dump(value)))
        elif isinstance(value, Substitution):
            normalized[tuple(name)] = tuple(normalize_to_list_of_substitutions(value))
        elif isinstance(value, (float, bool, int)):
            # Keep some types as is
            normalized[tuple(name)] = value
        elif isinstance(value, bytes):
            # Keep bytes as is
            normalized[tuple(name)] = value
        elif isinstance(value, Sequence):
            # try to make the parameter types uniform
            normalized[tuple(name)] = _normalize_parameter_array_value(value)
        else:
            raise TypeError('Unexpected type for parameter value {}'.format(repr(value)))
    return normalized


def normalize_parameters(parameters: SomeParameters) -> Parameters:
    """
    Normalize the types used to store parameters to substitution types.

    The passed parameters must be an iterable where each element is
    a path to a yaml file or a dict.
    The normalized parameters will have all paths converted to a list of :class:`Substitution`,
    and dictionaries normalized using :meth:`normalize_parameter_dict`.
    """
    if isinstance(parameters, str) or not isinstance(parameters, Sequence):
        raise TypeError('Expecting list of parameters, got {}'.format(parameters))

    normalized_params: List[Union[ParameterFile, ParametersDict, ParameterDescription]] = []
    for param in parameters:
        if isinstance(param, Mapping):
            normalized_params.append(normalize_parameter_dict(param))
        elif isinstance(param, ParameterDescription):
            normalized_params.append(param)
        elif isinstance(param, ParameterFile):
            normalized_params.append(param)
        else:
            # It's a path
            normalized_params.append(ParameterFile(param))
    return tuple(normalized_params)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for descriptions of launchable entities.

Descriptions are not executable and are immutable so they can be reused by launch entities.
"""

from .evaluate_parameters import evaluate_parameters
from .namespace_utils import is_namespace_absolute
from .namespace_utils import is_root_namespace
from .namespace_utils import make_namespace_absolute
from .namespace_utils import prefix_namespace
from .normalize_parameters import normalize_parameters
from .normalize_remap_rule import normalize_remap_rule
from .normalize_remap_rule import normalize_remap_rules
from .to_parameters_list import to_parameters_list
from .track_node_names import add_node_name
from .track_node_names import get_node_name_count

__all__ = [
    'add_node_name',
    'evaluate_parameters',
    'evaluate_parameters_dict',
    'get_node_name_count',
    'is_namespace_absolute',
    'is_root_namespace',
    'make_namespace_absolute',
    'normalize_parameters',
    'normalize_parameters_dict',
    'normalize_remap_rule',
    'normalize_remap_rules',
    'prefix_namespace',
    'to_parameters_list',
]



# Copyright 2018-2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing ROS specific adapters and their initialization."""

import os
import threading
from typing import List
from typing import Optional

import launch
import launch.events

import rclpy
from rclpy.executors import SingleThreadedExecutor


class ROSAdapter:
    """Wraps rclpy API to ease ROS node usage in `launch_ros` actions."""

    def __init__(
        self,
        *,
        argv: Optional[List[str]] = None,
        autostart: bool = True
    ):
        """
        Construct adapter.

        :param: argv List of global arguments for rclpy context initialization.
        :param: autostart Whether to start adapter on construction or not.
        """
        # Do not use `None` here, as `rclpy.init` will use `sys.argv` in that case.
        self.__argv = [] if argv is None else argv
        self.__ros_context = None
        self.__ros_node = None
        self.__ros_executor = None
        self.__is_running = False

        if autostart:
            self.start()

    def start(self):
        """Start ROS adapter."""
        if self.__is_running:
            raise RuntimeError('Cannot start a ROS adapter that is already running')
        self.__ros_context = rclpy.Context()
        rclpy.init(args=self.__argv, context=self.__ros_context)
        self.__ros_node = rclpy.create_node(
            'launch_ros_{}'.format(os.getpid()),
            context=self.__ros_context
        )
        self.__ros_executor = SingleThreadedExecutor(context=self.__ros_context)
        self.__is_running = True

        self.__ros_executor_thread = threading.Thread(target=self._run)
        self.__ros_executor_thread.start()

    def _run(self):
        try:
            self.__ros_executor.add_node(self.__ros_node)
            while self.__is_running:
                # TODO(wjwwood): switch this to `spin()` when it considers
                #   asynchronously added subscriptions.
                #   see: https://github.com/ros2/rclpy/issues/188
                self.__ros_executor.spin_once(timeout_sec=1.0)
        except KeyboardInterrupt:
            pass
        finally:
            self.__ros_executor.remove_node(self.__ros_node)

    def shutdown(self):
        """Shutdown ROS adapter."""
        if not self.__is_running:
            raise RuntimeError('Cannot shutdown a ROS adapter that is not running')
        self.__is_running = False
        self.__ros_executor_thread.join()
        self.__ros_node.destroy_node()
        rclpy.shutdown(context=self.__ros_context)

    @property
    def argv(self):
        return self.__argv

    @property
    def ros_context(self):
        return self.__ros_context

    @property
    def ros_node(self):
        return self.__ros_node

    @property
    def ros_executor(self):
        return self.__ros_executor


def get_ros_adapter(context: launch.LaunchContext):
    """
    Get the ROS adapter managed by the given launch context.

    If no adapter is found, one will be created.

    This function is reentrant but concurrent calls on the
    same `context` are not safe.
    """
    if not hasattr(context.locals, 'ros_adapter'):
        ros_adapter = ROSAdapter()
        context.extend_globals({'ros_adapter': ros_adapter})
        context.register_event_handler(launch.event_handlers.OnShutdown(
            on_shutdown=lambda *args, **kwargs: ros_adapter.shutdown()
        ))
    return context.locals.ros_adapter


def get_ros_node(context: launch.LaunchContext):
    """
    Get the ROS node managed by the given launch context.

    If no node is found, one will be created.

    This function is reentrant but concurrent calls on the
    same `context` are not safe.
    """
    return get_ros_adapter(context).ros_node



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the ExecutableInPackage substitution."""

import os
from typing import Iterable
from typing import List
from typing import Text

from launch.frontend import expose_substitution
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitution import Substitution
from launch.substitutions.substitution_failure import SubstitutionFailure
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions

from osrf_pycommon.process_utils import which

from .find_package import FindPackagePrefix


@expose_substitution('exec-in-pkg')
class ExecutableInPackage(FindPackagePrefix):
    """
    Substitution that tries to locate an executable in the libexec directory of a ROS package.

    The ROS package is located using ament_index_python.

    :raise: ament_index_python.packages.PackageNotFoundError when package is
        not found during substitution
    :raise: SubstitutionFailure when executable is not found, or package
        libexec directory doesn't exist, during substitution
    """

    def __init__(self, executable: SomeSubstitutionsType, package: SomeSubstitutionsType) -> None:
        """Create an ExecutableInPackage substitution."""
        super().__init__(package)
        self.__executable = normalize_to_list_of_substitutions(executable)

    @classmethod
    def parse(cls, data: Iterable[SomeSubstitutionsType]):
        """Parse a ExecutableInPackage substitution."""
        if not data or len(data) != 2:
            raise AttributeError('exec-in-package substitution expects 2 arguments')
        kwargs = {'executable': data[0], 'package': data[1]}
        return cls, kwargs

    @property
    def executable(self) -> List[Substitution]:
        """Getter for executable."""
        return self.__executable

    def describe(self) -> Text:
        """Return a description of this substitution as a string."""
        exec_str = ' + '.join([sub.describe() for sub in self.executable])
        pkg_str = ' + '.join([sub.describe() for sub in self.package])
        return 'ExecInPkg(pkg={}, exec={})'.format(pkg_str, exec_str)

    def perform(self, context: LaunchContext) -> Text:
        """Perform the substitution by locating the executable."""
        executable = perform_substitutions(context, self.executable)
        package = perform_substitutions(context, self.package)
        package_prefix = super().perform(context)
        package_libexec = os.path.join(package_prefix, 'lib', package)
        if not os.path.exists(package_libexec):
            raise SubstitutionFailure(
                "package '{}' found at '{}', but libexec directory '{}' does not exist".format(
                    package, package_prefix, package_libexec))
        result = which(executable, path=package_libexec)
        if result is None:
            raise SubstitutionFailure(
                "executable '{}' not found on the libexec directory '{}' ".format(
                    executable, package_libexec))
        return result



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the `SetParametersFromFile` action."""

from launch import Action
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions


@expose_action('set_parameters_from_file')
class SetParametersFromFile(Action):
    """
    Action that sets parameters for all nodes in scope based on a given yaml file.

    e.g.
    ```python3
        LaunchDescription([
            ...,
            GroupAction(
                actions = [
                    ...,
                    SetParametersFromFile('path/to/file.yaml'),
                    ...,
                    Node(...),  // the params will be passed to this node
                    ...,
                ]
            ),
            Node(...),  // here it won't be passed, as it's not in the same scope
            ...
        ])
    ```
    ```xml
    <launch>
        <group>
            <set_parameters_from_file filename='/path/to/file.yaml'/>
            <node .../>    // Node in scope, params will be passed
        </group>
        <node .../>  // Node not in scope, params won't be passed
    </launch>

    ```
    """

    def __init__(
        self,
        filename: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """Create a SetParameterFromFile action."""
        super().__init__(**kwargs)
        self._input_file = normalize_to_list_of_substitutions(filename)

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `SetParameterFromFile` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['filename'] = parser.parse_substitution(entity.get_attr('filename'))
        return cls, kwargs

    def execute(self, context: LaunchContext):
        """Execute the action."""
        filename = perform_substitutions(context, self._input_file)
        global_param_list = context.launch_configurations.get('global_params', [])
        global_param_list.append(filename)
        context.launch_configurations['global_params'] = global_param_list



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the LoadComposableNodes action."""

from pathlib import Path
import threading

from typing import List
from typing import Optional
from typing import Text
from typing import Union

import composition_interfaces.srv

from launch.action import Action
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
import launch.logging
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.some_substitutions_type import SomeSubstitutionsType_types_tuple
from launch.utilities import ensure_argument_type
from launch.utilities import is_a_subclass
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions
from launch_ros.parameter_descriptions import ParameterFile

from .composable_node_container import ComposableNodeContainer

from ..descriptions import ComposableNode
from ..ros_adapters import get_ros_node
from ..utilities import add_node_name
from ..utilities import evaluate_parameters
from ..utilities import get_node_name_count
from ..utilities import make_namespace_absolute
from ..utilities import prefix_namespace
from ..utilities import to_parameters_list
from ..utilities.normalize_parameters import normalize_parameter_dict


@expose_action('load_composable_node')
class LoadComposableNodes(Action):
    """Action that loads composable ROS nodes into a running container."""

    def __init__(
        self,
        *,
        composable_node_descriptions: List[ComposableNode],
        target_container: Union[SomeSubstitutionsType, ComposableNodeContainer],
        **kwargs,
    ) -> None:
        """
        Construct a LoadComposableNodes action.

        The container node is expected to provide a `~/_container/load_node` service for
        loading purposes.
        Loading will be performed sequentially.
        When executed, this action will block until the container's load service is available.
        Make sure any LoadComposableNode action is executed only after its container processes
        has started.

        :param composable_node_descriptions: descriptions of composable nodes to be loaded
        :param target_container: the container to load the nodes into
        """
        ensure_argument_type(
            target_container,
            list(SomeSubstitutionsType_types_tuple) +
            [ComposableNodeContainer],
            'target_container',
            'LoadComposableNodes'
        )
        super().__init__(**kwargs)
        self.__composable_node_descriptions = composable_node_descriptions
        self.__target_container = target_container
        self.__final_target_container_name = None  # type: Optional[Text]
        self.__logger = launch.logging.get_logger(__name__)

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Parse load_composable_node."""
        _, kwargs = super().parse(entity, parser)

        kwargs['target_container'] = parser.parse_substitution(
            entity.get_attr('target', data_type=str))

        composable_nodes = entity.get_attr('composable_node', data_type=List[Entity])
        kwargs['composable_node_descriptions'] = []
        for entity in composable_nodes:
            composable_node_cls, composable_node_kwargs = ComposableNode.parse(parser, entity)
            kwargs['composable_node_descriptions'].append(
                composable_node_cls(**composable_node_kwargs))

        return cls, kwargs

    def _load_node(
        self,
        request: composition_interfaces.srv.LoadNode.Request,
        context: LaunchContext
    ) -> None:
        """
        Load node synchronously.

        :param request: service request to load a node
        :param context: current launch context
        """
        while not self.__rclpy_load_node_client.wait_for_service(timeout_sec=1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service, due to shutdown.".format(
                        self.__rclpy_load_node_client.srv_name
                    )
                )
                return

        # Asynchronously wait on service call so that we can periodically check for shutdown
        event = threading.Event()

        def unblock(future):
            nonlocal event
            event.set()

        self.__logger.debug(
            "Calling the '{}' service with request '{}'".format(
                self.__rclpy_load_node_client.srv_name, request
            )
        )

        response_future = self.__rclpy_load_node_client.call_async(request)
        response_future.add_done_callback(unblock)

        while not event.wait(1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service response, due to shutdown.".format(
                        self.__rclpy_load_node_client.srv_name),
                )
                response_future.cancel()
                return

        # Get response
        if response_future.exception() is not None:
            raise response_future.exception()
        response = response_future.result()

        self.__logger.debug("Received response '{}'".format(response))

        node_name = response.full_node_name if response.full_node_name else request.node_name
        if response.success:
            if node_name is not None:
                add_node_name(context, node_name)
                node_name_count = get_node_name_count(context, node_name)
                if node_name_count > 1:
                    container_logger = launch.logging.get_logger(
                        self.__final_target_container_name
                    )
                    container_logger.warning(
                        'there are now at least {} nodes with the name {} created within this '
                        'launch context'.format(node_name_count, node_name)
                    )
            self.__logger.info("Loaded node '{}' in container '{}'".format(
                response.full_node_name, self.__final_target_container_name
            ))
        else:
            self.__logger.error(
                "Failed to load node '{}' of type '{}' in container '{}': {}".format(
                    node_name, request.plugin_name, self.__final_target_container_name,
                    response.error_message
                )
            )

    def _load_in_sequence(
        self,
        load_node_requests: List[composition_interfaces.srv.LoadNode.Request],
        context: LaunchContext
    ) -> None:
        """
        Load composable nodes sequentially.

        :param load_node_requests: a list of LoadNode service requests to execute
        :param context: current launch context
        """
        next_load_node_request = load_node_requests[0]
        load_node_requests = load_node_requests[1:]
        self._load_node(next_load_node_request, context)
        if len(load_node_requests) > 0:
            context.add_completion_future(
                context.asyncio_loop.run_in_executor(
                    None, self._load_in_sequence, load_node_requests, context
                )
            )

    def execute(
        self,
        context: LaunchContext
    ) -> Optional[List[Action]]:
        """Execute the action."""
        # resolve target container node name

        if is_a_subclass(self.__target_container, ComposableNodeContainer):
            self.__final_target_container_name = self.__target_container.node_name
        elif isinstance(self.__target_container, SomeSubstitutionsType_types_tuple):
            subs = normalize_to_list_of_substitutions(self.__target_container)
            self.__final_target_container_name = perform_substitutions(
                context, subs)
        else:
            self.__logger.error(
                'target container is neither a ComposableNodeContainer nor a SubstitutionType')
            return

        # Create a client to load nodes in the target container.
        self.__rclpy_load_node_client = get_ros_node(context).create_client(
            composition_interfaces.srv.LoadNode, '{}/_container/load_node'.format(
                self.__final_target_container_name
            )
        )

        # Generate load requests before execute() exits to avoid race with context changing
        # due to scope change (e.g. if loading nodes from within a GroupAction).
        load_node_requests = [
            get_composable_node_load_request(node_description, context)
            for node_description in self.__composable_node_descriptions
        ]

        context.add_completion_future(
            context.asyncio_loop.run_in_executor(
                None, self._load_in_sequence, load_node_requests, context
            )
        )


def get_composable_node_load_request(
    composable_node_description: ComposableNode,
    context: LaunchContext
):
    """Get the request that will be sent to the composable node container."""
    request = composition_interfaces.srv.LoadNode.Request()
    request.package_name = perform_substitutions(
        context, composable_node_description.package
    )
    request.plugin_name = perform_substitutions(
        context, composable_node_description.node_plugin
    )
    if composable_node_description.node_name is not None:
        request.node_name = perform_substitutions(
            context, composable_node_description.node_name
        )
    expanded_ns = composable_node_description.node_namespace
    if expanded_ns is not None:
        expanded_ns = perform_substitutions(context, expanded_ns)
    base_ns = context.launch_configurations.get('ros_namespace', None)
    combined_ns = make_namespace_absolute(prefix_namespace(base_ns, expanded_ns))
    if combined_ns is not None:
        request.node_namespace = combined_ns
    # request.log_level = perform_substitutions(context, node_description.log_level)
    remappings = []
    global_remaps = context.launch_configurations.get('ros_remaps', None)
    if global_remaps:
        remappings.extend([f'{src}:={dst}' for src, dst in global_remaps])
    if composable_node_description.remappings:
        remappings.extend([
            f'{perform_substitutions(context, src)}:={perform_substitutions(context, dst)}'
            for src, dst in composable_node_description.remappings
        ])
    if remappings:
        request.remap_rules = remappings
    params_container = context.launch_configurations.get('global_params', None)
    parameters = []
    if params_container is not None:
        for param in params_container:
            if isinstance(param, tuple):
                subs = normalize_parameter_dict({param[0]: param[1]})
                parameters.append(subs)
            else:
                param_file_path = Path(param).resolve()
                assert param_file_path.is_file()
                subs = ParameterFile(param_file_path)
                parameters.append(subs)
    if composable_node_description.parameters is not None:
        parameters.extend(list(composable_node_description.parameters))
    if parameters:
        request.parameters = [
            param.to_parameter_msg() for param in to_parameters_list(
                context, request.node_name, combined_ns,
                evaluate_parameters(
                    context, parameters
                )
            )
        ]

    if composable_node_description.extra_arguments is not None:
        request.extra_arguments = [
            param.to_parameter_msg() for param in to_parameters_list(
                context, request.node_name, combined_ns,
                evaluate_parameters(
                    context, composable_node_description.extra_arguments
                )
            )
        ]
    return request



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the Node action."""

import os
import pathlib
from tempfile import NamedTemporaryFile
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Text  # noqa: F401
from typing import Tuple  # noqa: F401
from typing import Union

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

from launch.action import Action
from launch.actions import ExecuteProcess
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.frontend.type_utils import get_data_type_from_identifier

from launch.launch_context import LaunchContext
import launch.logging
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitutions import LocalSubstitution
from launch.utilities import ensure_argument_type
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions

from launch_ros.parameters_type import SomeParameters
from launch_ros.remap_rule_type import SomeRemapRules
from launch_ros.substitutions import ExecutableInPackage
from launch_ros.utilities import add_node_name
from launch_ros.utilities import evaluate_parameters
from launch_ros.utilities import get_node_name_count
from launch_ros.utilities import make_namespace_absolute
from launch_ros.utilities import normalize_parameters
from launch_ros.utilities import normalize_remap_rules
from launch_ros.utilities import plugin_support
from launch_ros.utilities import prefix_namespace


from rclpy.validate_namespace import validate_namespace
from rclpy.validate_node_name import validate_node_name

import yaml

from ..descriptions import Parameter
from ..descriptions import ParameterFile


class NodeActionExtension:
    """
    The extension point for launch_ros node action extensions.

    The following properties must be defined:
    * `NAME` (will be set to the entry point name)

    The following methods may be defined:
    * `command_extension`
    * `execute`
    """

    NAME = None
    EXTENSION_POINT_VERSION = '0.1'

    def __init__(self):
        super(NodeActionExtension, self).__init__()
        plugin_support.satisfies_version(self.EXTENSION_POINT_VERSION, '^0.1')

    def prepare_for_execute(self, context, ros_specific_arguments, node_action):
        """
        Perform any actions prior to the node's process being launched.

        `context` is the context within which the launch is taking place,
        containing amongst other things the command line arguments provided by
        the user.

        `ros_specific_arguments` is a dictionary of command line arguments that
        will be passed to the executable and are specific to ROS.

        `node_action` is the Node action instance that is calling the
        extension.

        This method must return a tuple of command line additions as a list of
        launch.substitutions.TextSubstitution objects, and
        `ros_specific_arguments` with any modifications made to it. If no
        modifications are made, it should return
        `[], ros_specific_arguments`.
        """
        return [], ros_specific_arguments


@expose_action('node')
class Node(ExecuteProcess):
    """Action that executes a ROS node."""

    UNSPECIFIED_NODE_NAME = '<node_name_unspecified>'
    UNSPECIFIED_NODE_NAMESPACE = '<node_namespace_unspecified>'

    def __init__(
        self, *,
        executable: SomeSubstitutionsType,
        package: Optional[SomeSubstitutionsType] = None,
        name: Optional[SomeSubstitutionsType] = None,
        namespace: Optional[SomeSubstitutionsType] = None,
        exec_name: Optional[SomeSubstitutionsType] = None,
        parameters: Optional[SomeParameters] = None,
        remappings: Optional[SomeRemapRules] = None,
        ros_arguments: Optional[Iterable[SomeSubstitutionsType]] = None,
        arguments: Optional[Iterable[SomeSubstitutionsType]] = None,
        **kwargs
    ) -> None:
        """
        Construct an Node action.

        Many arguments are passed eventually to
        :class:`launch.actions.ExecuteProcess`, so see the documentation of
        that class for additional details.
        However, the `cmd` is not meant to be used, instead use the
        `executable` and `arguments` keyword arguments to this function.

        This action, once executed, delegates most work to the
        :class:`launch.actions.ExecuteProcess`, but it also converts some ROS
        specific arguments into generic command line arguments.

        The launch_ros.substitutions.ExecutableInPackage substitution is used
        to find the executable at runtime, so this Action also raise the
        exceptions that substitution can raise when the package or executable
        are not found.

        If the name is not given (or is None) then no name is passed to
        the node on creation and instead the default name specified within the
        code of the node is used instead.

        The namespace can either be absolute (i.e. starts with /) or
        relative.
        If absolute, then nothing else is considered and this is passed
        directly to the node to set the namespace.
        If relative, the namespace in the 'ros_namespace' LaunchConfiguration
        will be prepended to the given relative node namespace.
        If no namespace is given, then the default namespace `/` is
        assumed.

        The parameters are passed as a list, with each element either a yaml
        file that contains parameter rules (string or pathlib.Path to the full
        path of the file), or a dictionary that specifies parameter rules.
        Keys of the dictionary can be strings or an iterable of Substitutions
        that will be expanded to a string.
        Values in the dictionary can be strings, integers, floats, or tuples
        of Substitutions that will be expanded to a string.
        Additionally, values in the dictionary can be lists of the
        aforementioned types, or another dictionary with the same properties.
        A yaml file with the resulting parameters from the dictionary will be
        written to a temporary file, the path to which will be passed to the
        node.
        Multiple parameter dictionaries/files can be passed: each file path
        will be passed in in order to the node (where the last definition of
        a parameter takes effect).
        However, fully qualified node names override wildcards even when
        specified earlier.
        If `namespace` is not specified, dictionaries are prefixed by a
        wildcard namespace (`/**`) and other specific parameter declarations
        may overwrite it.

        Using `ros_arguments` is equivalent to using `arguments` with a
        prepended '--ros-args' item.

        :param: executable the name of the executable to find if a package
            is provided or otherwise a path to the executable to run.
        :param: package the package in which the node executable can be found
        :param: name the name of the node
        :param: namespace the ROS namespace for this Node
        :param: exec_name the label used to represent the process.
            Defaults to the basename of node executable.
        :param: parameters list of names of yaml files with parameter rules,
            or dictionaries of parameters.
        :param: remappings ordered list of 'to' and 'from' string pairs to be
            passed to the node as ROS remapping rules
        :param: ros_arguments list of ROS arguments for the node
        :param: arguments list of extra arguments for the node
        """
        if package is not None:
            cmd = [ExecutableInPackage(package=package, executable=executable)]
        else:
            cmd = [executable]
        cmd += [] if arguments is None else arguments
        cmd += [] if ros_arguments is None else ['--ros-args'] + ros_arguments
        # Reserve space for ros specific arguments.
        # The substitutions will get expanded when the action is executed.
        cmd += ['--ros-args']  # Prepend ros specific arguments with --ros-args flag
        if name is not None:
            cmd += ['-r', LocalSubstitution(
                "ros_specific_arguments['name']", description='node name')]
        if parameters is not None:
            ensure_argument_type(parameters, (list), 'parameters', 'Node')
            # All elements in the list are paths to files with parameters (or substitutions that
            # evaluate to paths), or dictionaries of parameters (fields can be substitutions).
            normalized_params = normalize_parameters(parameters)
        # Forward 'exec_name' as to ExecuteProcess constructor
        kwargs['name'] = exec_name
        super().__init__(cmd=cmd, **kwargs)
        self.__package = package
        self.__node_executable = executable
        self.__node_name = name
        self.__node_namespace = namespace
        self.__parameters = [] if parameters is None else normalized_params
        self.__remappings = [] if remappings is None else list(normalize_remap_rules(remappings))
        self.__ros_arguments = ros_arguments
        self.__arguments = arguments

        self.__expanded_node_name = self.UNSPECIFIED_NODE_NAME
        self.__expanded_node_namespace = self.UNSPECIFIED_NODE_NAMESPACE
        self.__expanded_parameter_arguments = None  # type: Optional[List[Tuple[Text, bool]]]
        self.__final_node_name = None  # type: Optional[Text]
        self.__expanded_remappings = None  # type: Optional[List[Tuple[Text, Text]]]

        self.__substitutions_performed = False

        self.__logger = launch.logging.get_logger(__name__)

        self.__extensions = get_extensions(self.__logger)

    @staticmethod
    def parse_nested_parameters(params, parser):
        """Normalize parameters as expected by Node constructor argument."""
        from ..descriptions import ParameterValue

        def get_nested_dictionary_from_nested_key_value_pairs(params):
            """Convert nested params in a nested dictionary."""
            param_dict = {}
            for param in params:
                name = tuple(parser.parse_substitution(param.get_attr('name')))
                type_identifier = param.get_attr('type', data_type=None, optional=True)
                data_type = None
                if type_identifier is not None:
                    data_type = get_data_type_from_identifier(type_identifier)
                value = param.get_attr('value', data_type=data_type, optional=True)
                nested_params = param.get_attr('param', data_type=List[Entity], optional=True)
                param.assert_entity_completely_parsed()
                if value is not None and nested_params:
                    raise RuntimeError(
                        'nested parameters and value attributes are mutually exclusive')
                if data_type is not None and nested_params:
                    raise RuntimeError(
                        'nested parameters and type attributes are mutually exclusive')
                elif value is not None:
                    some_value = parser.parse_if_substitutions(value)
                    param_dict[name] = ParameterValue(some_value, value_type=data_type)
                elif nested_params:
                    param_dict.update({
                        name: get_nested_dictionary_from_nested_key_value_pairs(nested_params)
                    })
                else:
                    raise RuntimeError('either a value attribute or nested params are needed')
            return param_dict

        normalized_params = []
        for param in params:
            from_attr = param.get_attr('from', optional=True)
            allow_substs = param.get_attr('allow_substs', data_type=bool, optional=True)
            name = param.get_attr('name', optional=True)
            if from_attr is not None and name is not None:
                raise RuntimeError('name and from attributes are mutually exclusive')
            elif from_attr is not None:
                # 'from' attribute ignores 'name' attribute,
                # it's not accepted to be nested,
                # and it can not have children.
                if isinstance(allow_substs, str):
                    allow_substs = parser.parse_substitution(allow_substs)
                else:
                    allow_substs = bool(allow_substs)
                param.assert_entity_completely_parsed()
                normalized_params.append(
                    ParameterFile(parser.parse_substitution(from_attr), allow_substs=allow_substs))
                continue
            elif name is not None:
                if allow_substs is not None:
                    raise RuntimeError(
                        "'allow_substs' can only be used together with 'from' attribute")
                normalized_params.append(
                    get_nested_dictionary_from_nested_key_value_pairs([param]))
                continue
            raise ValueError('param Entity should have name or from attribute')
        return normalized_params

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Parse node."""
        # See parse method of `ExecuteProcess`
        _, kwargs = super().parse(entity, parser, ignore=['cmd'])
        args = entity.get_attr('args', optional=True)
        if args is not None:
            kwargs['arguments'] = super()._parse_cmdline(args, parser)
        ros_args = entity.get_attr('ros_args', optional=True)
        if ros_args is not None:
            kwargs['ros_arguments'] = super()._parse_cmdline(ros_args, parser)
        node_name = entity.get_attr('node-name', optional=True)
        if node_name is not None:
            kwargs['node_name'] = parser.parse_substitution(node_name)
        node_name = entity.get_attr('name', optional=True)
        if node_name is not None:
            kwargs['name'] = parser.parse_substitution(node_name)
        exec_name = entity.get_attr('exec_name', optional=True)
        if exec_name is not None:
            kwargs['exec_name'] = parser.parse_substitution(exec_name)
        package = entity.get_attr('pkg', optional=True)
        if package is not None:
            kwargs['package'] = parser.parse_substitution(package)
        kwargs['executable'] = parser.parse_substitution(entity.get_attr('exec'))
        ns = entity.get_attr('namespace', optional=True)
        if ns is not None:
            kwargs['namespace'] = parser.parse_substitution(ns)
        remappings = entity.get_attr('remap', data_type=List[Entity], optional=True)
        if remappings is not None:
            kwargs['remappings'] = [
                (
                    parser.parse_substitution(remap.get_attr('from')),
                    parser.parse_substitution(remap.get_attr('to'))
                ) for remap in remappings
            ]
            for remap in remappings:
                remap.assert_entity_completely_parsed()
        parameters = entity.get_attr('param', data_type=List[Entity], optional=True)
        if parameters is not None:
            kwargs['parameters'] = cls.parse_nested_parameters(parameters, parser)

        return cls, kwargs

    @property
    def node_package(self):
        """Getter for node_package."""
        return self.__package

    @property
    def node_executable(self):
        """Getter for node_executable."""
        return self.__node_executable

    @property
    def node_name(self):
        """Getter for node_name."""
        if self.__final_node_name is None:
            raise RuntimeError("cannot access 'node_name' before executing action")
        return self.__final_node_name

    def is_node_name_fully_specified(self):
        keywords = (self.UNSPECIFIED_NODE_NAME, self.UNSPECIFIED_NODE_NAMESPACE)
        return all(x not in self.node_name for x in keywords)

    def _create_params_file_from_dict(self, params):
        with NamedTemporaryFile(mode='w', prefix='launch_params_', delete=False) as h:
            param_file_path = h.name
            param_dict = {
                self.node_name if self.is_node_name_fully_specified() else '/**':
                {'ros__parameters': params}
            }
            yaml.dump(param_dict, h, default_flow_style=False)
            return param_file_path

    def _get_parameter_rule(self, param: 'Parameter', context: LaunchContext):
        name, value = param.evaluate(context)
        return f'{name}:={yaml.dump(value)}'

    def _perform_substitutions(self, context: LaunchContext) -> None:
        # Here to avoid cyclic import
        from ..descriptions import Parameter
        try:
            if self.__substitutions_performed:
                # This function may have already been called by a subclass' `execute`, for example.
                return
            self.__substitutions_performed = True
            if self.__node_name is not None:
                self.__expanded_node_name = perform_substitutions(
                    context, normalize_to_list_of_substitutions(self.__node_name))
                validate_node_name(self.__expanded_node_name)
            self.__expanded_node_name.lstrip('/')
            expanded_node_namespace: Optional[Text] = None
            if self.__node_namespace is not None:
                expanded_node_namespace = perform_substitutions(
                    context, normalize_to_list_of_substitutions(self.__node_namespace))
            base_ns = context.launch_configurations.get('ros_namespace', None)
            expanded_node_namespace = make_namespace_absolute(
                prefix_namespace(base_ns, expanded_node_namespace))
            if expanded_node_namespace is not None:
                self.__expanded_node_namespace = expanded_node_namespace
                cmd_extension = ['-r', LocalSubstitution("ros_specific_arguments['ns']")]
                self.cmd.extend([normalize_to_list_of_substitutions(x) for x in cmd_extension])
                validate_namespace(self.__expanded_node_namespace)
        except Exception:
            self.__logger.error(
                "Error while expanding or validating node name or namespace for '{}':"
                .format('package={}, executable={}, name={}, namespace={}'.format(
                    self.__package,
                    self.__node_executable,
                    self.__node_name,
                    self.__node_namespace,
                ))
            )
            raise
        self.__final_node_name = prefix_namespace(
            self.__expanded_node_namespace, self.__expanded_node_name)

        # Expand global parameters first,
        # so they can be overridden with specific parameters of this Node
        # The params_container list is expected to contain name-value pairs (tuples)
        # and/or strings representing paths to parameter files.
        params_container = context.launch_configurations.get('global_params', None)

        if any(x is not None for x in (params_container, self.__parameters)):
            self.__expanded_parameter_arguments = []
        if params_container is not None:
            for param in params_container:
                if isinstance(param, tuple):
                    name, value = param
                    cmd_extension = ['-p', f'{name}:={value}']
                    self.cmd.extend([normalize_to_list_of_substitutions(x) for x in cmd_extension])
                else:
                    param_file_path = os.path.abspath(param)
                    self.__expanded_parameter_arguments.append((param_file_path, True))
                    cmd_extension = ['--params-file', f'{param_file_path}']
                    assert os.path.isfile(param_file_path)
                    self.cmd.extend([normalize_to_list_of_substitutions(x) for x in cmd_extension])

        # expand parameters too
        if self.__parameters is not None:
            evaluated_parameters = evaluate_parameters(context, self.__parameters)
            for params in evaluated_parameters:
                is_file = False
                if isinstance(params, dict):
                    param_argument = self._create_params_file_from_dict(params)
                    is_file = True
                    assert os.path.isfile(param_argument)
                elif isinstance(params, pathlib.Path):
                    param_argument = str(params)
                    is_file = True
                elif isinstance(params, Parameter):
                    param_argument = self._get_parameter_rule(params, context)
                else:
                    raise RuntimeError('invalid normalized parameters {}'.format(repr(params)))
                if is_file and not os.path.isfile(param_argument):
                    self.__logger.warning(
                        'Parameter file path is not a file: {}'.format(param_argument),
                    )
                    continue
                self.__expanded_parameter_arguments.append((param_argument, is_file))
                cmd_extension = ['--params-file' if is_file else '-p', f'{param_argument}']
                self.cmd.extend([normalize_to_list_of_substitutions(x) for x in cmd_extension])
        # expand remappings too
        global_remaps = context.launch_configurations.get('ros_remaps', None)
        if global_remaps or self.__remappings:
            self.__expanded_remappings = []
        if global_remaps:
            self.__expanded_remappings.extend(global_remaps)
        if self.__remappings:
            self.__expanded_remappings.extend([
                (perform_substitutions(context, src), perform_substitutions(context, dst))
                for src, dst in self.__remappings
            ])
        if self.__expanded_remappings:
            cmd_extension = []
            for src, dst in self.__expanded_remappings:
                cmd_extension.extend(['-r', f'{src}:={dst}'])
            self.cmd.extend([normalize_to_list_of_substitutions(x) for x in cmd_extension])

    def execute(self, context: LaunchContext) -> Optional[List[Action]]:
        """
        Execute the action.

        Delegated to :meth:`launch.actions.ExecuteProcess.execute`.
        """
        self._perform_substitutions(context)
        # Prepare the ros_specific_arguments list and add it to the context so that the
        # LocalSubstitution placeholders added to the the cmd can be expanded using the contents.
        ros_specific_arguments: Dict[str, Union[str, List[str]]] = {}
        if self.__node_name is not None:
            ros_specific_arguments['name'] = '__node:={}'.format(self.__expanded_node_name)
        if self.__expanded_node_namespace != '':
            ros_specific_arguments['ns'] = '__ns:={}'.format(self.__expanded_node_namespace)

        # Give extensions a chance to prepare for execution
        for extension in self.__extensions.values():
            cmd_extension, ros_specific_arguments = extension.prepare_for_execute(
                context,
                ros_specific_arguments,
                self
            )
            self.cmd.extend(cmd_extension)

        context.extend_locals({'ros_specific_arguments': ros_specific_arguments})
        ret = super().execute(context)

        if self.is_node_name_fully_specified():
            add_node_name(context, self.node_name)
            node_name_count = get_node_name_count(context, self.node_name)
            if node_name_count > 1:
                execute_process_logger = launch.logging.get_logger(self.name)
                execute_process_logger.warning(
                    'there are now at least {} nodes with the name {} created within this '
                    'launch context'.format(node_name_count, self.node_name)
                )

        return ret

    @property
    def expanded_node_namespace(self):
        """Getter for expanded_node_namespace."""
        return self.__expanded_node_namespace

    @property
    def expanded_remapping_rules(self):
        """Getter for expanded_remappings."""
        return self.__expanded_remappings


def instantiate_extension(
    group_name,
    extension_name,
    extension_class,
    extensions,
    logger,
    *,
    unique_instance=False
):
    if not unique_instance and extension_class in extensions:
        return extensions[extension_name]
    try:
        extension_instance = extension_class()
    except plugin_support.PluginException as e:  # noqa: F841
        logger.warning(
            f"Failed to instantiate '{group_name}' extension "
            f"'{extension_name}': {e}")
        return None
    except Exception as e:  # noqa: F841
        logger.error(
            f"Failed to instantiate '{group_name}' extension "
            f"'{extension_name}': {e}")
        return None
    if not unique_instance:
        extensions[extension_name] = extension_instance
    return extension_instance


def get_extensions(logger):
    group_name = 'launch_ros.node_action'
    entry_points_impl = importlib_metadata.entry_points()
    if hasattr(entry_points_impl, 'select'):
        groups = entry_points_impl.select(group=group_name)
    else:
        groups = entry_points_impl.get(group_name, [])
    entry_points = {}
    for entry_point in groups:
        entry_points[entry_point.name] = entry_point
    extension_types = {}
    for entry_point in entry_points:
        try:
            extension_type = entry_points[entry_point].load()
        except Exception as e:  # noqa: F841
            logger.warning(f"Failed to load entry point '{entry_points[entry_point].name}': {e}")
            continue
        extension_types[entry_points[entry_point].name] = extension_type

    extensions = {}
    for extension_name, extension_class in extension_types.items():
        extension_instance = instantiate_extension(
            group_name, extension_name, extension_class, extensions, logger)
        if extension_instance is None:
            continue
        extension_instance.NAME = extension_name
        extensions[extension_name] = extension_instance
    return extensions



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the LifecycleNode action."""

import functools
import threading
from typing import cast
from typing import List
from typing import Optional

import launch
from launch import SomeSubstitutionsType
from launch.action import Action
import launch.logging

import lifecycle_msgs.msg
import lifecycle_msgs.srv

from .node import Node
from ..events.lifecycle import ChangeState
from ..events.lifecycle import StateTransition

from ..ros_adapters import get_ros_node


class LifecycleNode(Node):
    """Action that executes a ROS lifecycle node."""

    def __init__(
        self,
        *,
        name: SomeSubstitutionsType,
        namespace: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """
        Construct a LifecycleNode action.

        Almost all of the arguments are passed to :class:`Node` and eventually
        to :class:`launch.actions.ExecuteProcess`, so see the documentation of
        those classes for additional details.

        This action additionally emits some event(s) in certain circumstances:

        - :class:`launch.events.lifecycle.StateTransition`:

            - this event is emitted when a message is published to the
              "/<name>/transition_event" topic, indicating the lifecycle
              node represented by this action changed state

        This action also handles some events related to lifecycle:

        - :class:`launch.events.lifecycle.ChangeState`

          - this event can be targeted to a single lifecycle node, or more than
            one, or even all lifecycle nodes, and it requests the targeted nodes
            to change state, see its documentation for more details.

        :param name: The name of the lifecycle node.
          Although it defaults to None it is a required parameter and the default will be removed
          in a future release.
        """
        super().__init__(name=name, namespace=namespace, **kwargs)
        self.__logger = launch.logging.get_logger(__name__)
        self.__rclpy_subscription = None
        self.__current_state = \
            ChangeState.valid_states[lifecycle_msgs.msg.State.PRIMARY_STATE_UNKNOWN]

    def _on_transition_event(self, context, msg):
        try:
            event = StateTransition(action=self, msg=msg)
            self.__current_state = ChangeState.valid_states[msg.goal_state.id]
            context.asyncio_loop.call_soon_threadsafe(lambda: context.emit_event_sync(event))
        except Exception as exc:
            self.__logger.error(
                "Exception in handling of 'lifecycle.msg.TransitionEvent': {}".format(exc))

    def _call_change_state(self, request, context: launch.LaunchContext):
        while not self.__rclpy_change_state_client.wait_for_service(timeout_sec=1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service, due to shutdown.".format(
                        self.__rclpy_change_state_client.srv_name),
                )
                return

        # Asynchronously wait so that we can periodically check for shutdown.
        event = threading.Event()

        def unblock(future):
            nonlocal event
            event.set()

        response_future = self.__rclpy_change_state_client.call_async(request)
        response_future.add_done_callback(unblock)

        while not event.wait(1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service response, due to shutdown.".format(
                        self.__rclpy_change_state_client.srv_name),
                )
                response_future.cancel()
                return

        if response_future.exception() is not None:
            raise response_future.exception()
        response = response_future.result()

        if not response.success:
            self.__logger.error(
                "Failed to make transition '{}' for LifecycleNode '{}'".format(
                    ChangeState.valid_transitions[request.transition.id],
                    self.node_name,
                )
            )

    def _on_change_state_event(self, context: launch.LaunchContext) -> None:
        typed_event = cast(ChangeState, context.locals.event)
        if not typed_event.lifecycle_node_matcher(self):
            return None
        request = lifecycle_msgs.srv.ChangeState.Request()
        request.transition.id = typed_event.transition_id
        context.add_completion_future(
            context.asyncio_loop.run_in_executor(None, self._call_change_state, request, context))

    def execute(self, context: launch.LaunchContext) -> Optional[List[Action]]:
        """
        Execute the action.

        Delegated to :meth:`launch.actions.ExecuteProcess.execute`.
        """
        self._perform_substitutions(context)  # ensure self.node_name is expanded
        if '<node_name_unspecified>' in self.node_name:
            raise RuntimeError('node_name unexpectedly incomplete for lifecycle node')
        node = get_ros_node(context)
        # Create a subscription to monitor the state changes of the subprocess.
        self.__rclpy_subscription = node.create_subscription(
            lifecycle_msgs.msg.TransitionEvent,
            '{}/transition_event'.format(self.node_name),
            functools.partial(self._on_transition_event, context),
            10)
        # Create a service client to change state on demand.
        self.__rclpy_change_state_client = node.create_client(
            lifecycle_msgs.srv.ChangeState,
            '{}/change_state'.format(self.node_name))
        # Register an event handler to change states on a ChangeState lifecycle event.
        context.register_event_handler(launch.EventHandler(
            matcher=lambda event: isinstance(event, ChangeState),
            entities=[launch.actions.OpaqueFunction(function=self._on_change_state_event)],
        ))
        # Delegate execution to Node and ExecuteProcess.
        return super().execute(context)



# Copyright 2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the `SetRemap` action."""

from typing import List

from launch import Action
from launch import Substitution
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions


@expose_action('set_remap')
class SetRemap(Action):
    """
    Action that sets a remapping rule in the current context.

    This remapping rule will be passed to all the nodes launched in the same scope, overriding
    the ones specified in the `Node` action constructor.
    e.g.:
    ```python3
        LaunchDescription([
            ...,
            GroupAction(
                actions = [
                    ...,
                    SetRemap(src='asd', dst='bsd'),
                    ...,
                    Node(...),  // the remap rule will be passed to this node
                    ...,
                ]
            ),
            Node(...),  // here it won't be passed, as it's not in the same scope
            ...
        ])
    ```
    """

    def __init__(
        self,
        src: SomeSubstitutionsType,
        dst: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """Create a SetRemap action."""
        super().__init__(**kwargs)
        self.__src = normalize_to_list_of_substitutions(src)
        self.__dst = normalize_to_list_of_substitutions(dst)

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `SetRemap` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['src'] = parser.parse_substitution(entity.get_attr('from'))
        kwargs['dst'] = parser.parse_substitution(entity.get_attr('to'))
        return cls, kwargs

    @property
    def src(self) -> List[Substitution]:
        """Getter for src."""
        return self.__src

    @property
    def dst(self) -> List[Substitution]:
        """Getter for dst."""
        return self.__dst

    def execute(self, context: LaunchContext):
        """Execute the action."""
        src = perform_substitutions(context, self.__src)
        dst = perform_substitutions(context, self.__dst)
        global_remaps = context.launch_configurations.get('ros_remaps', [])
        global_remaps.append((src, dst))
        context.launch_configurations['ros_remaps'] = global_remaps



# Copyright 2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the `SetParameter` action."""

from launch import Action
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.utilities import normalize_to_list_of_substitutions

from launch_ros.parameters_type import ParameterName
from launch_ros.parameters_type import ParameterValue
from launch_ros.parameters_type import SomeParameterValue
from launch_ros.utilities.evaluate_parameters import evaluate_parameter_dict
from launch_ros.utilities.normalize_parameters import normalize_parameter_dict


@expose_action('set_parameter')
class SetParameter(Action):
    """
    Action that sets a parameter in the current context.

    This parameter will be set in all the nodes launched in the same scope.
    e.g.:
    ```python3
        LaunchDescription([
            ...,
            GroupAction(
                actions = [
                    ...,
                    SetParameter(name='my_param', value='2'),
                    ...,
                    Node(...),  // the param will be passed to this node
                    ...,
                ]
            ),
            Node(...),  // here it won't be passed, as it's not in the same scope
            ...
        ])
    ```
    """

    def __init__(
        self,
        name: SomeSubstitutionsType,
        value: SomeParameterValue,
        **kwargs
    ) -> None:
        """Create a SetParameter action."""
        super().__init__(**kwargs)
        normalized_name = normalize_to_list_of_substitutions(name)
        self.__param_dict = normalize_parameter_dict({tuple(normalized_name): value})

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `SetParameter` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['name'] = parser.parse_substitution(entity.get_attr('name'))
        kwargs['value'] = parser.parse_substitution(entity.get_attr('value'))
        return cls, kwargs

    @property
    def name(self) -> ParameterName:
        """Getter for name."""
        return list(self.__param_dict.keys())[0]

    @property
    def value(self) -> ParameterValue:
        """Getter for value."""
        return list(self.__param_dict.values())[0]

    def execute(self, context: LaunchContext):
        """Execute the action."""
        eval_param_dict = evaluate_parameter_dict(context, self.__param_dict)
        global_param_list = context.launch_configurations.get('global_params', [])
        global_param_list.extend(eval_param_dict.items())
        context.launch_configurations['global_params'] = global_param_list



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from packaging.version import Version


class PluginException(Exception):
    """Base class for all exceptions within the plugin system."""

    pass


def satisfies_version(version, caret_range):
    assert caret_range.startswith('^'), 'Only supports caret ranges'
    extension_point_version = Version(version)
    extension_version = Version(caret_range[1:])
    next_extension_version = get_upper_bound_caret_version(
        extension_version)

    if extension_point_version < extension_version:
        raise PluginException(
            'Extension point is too old (%s), the extension requires '
            "'%s'" % (extension_point_version, extension_version))

    if extension_point_version >= next_extension_version:
        raise PluginException(
            'Extension point is newer (%s), than what the extension '
            "supports '%s'" % (extension_point_version, extension_version))


def get_upper_bound_caret_version(version):
    parts = version.base_version.split('.')
    if len(parts) < 2:
        parts += [0] * (2 - len(parts))
    major, minor = [int(p) for p in parts[:2]]
    if major > 0:
        major += 1
        minor = 0
    else:
        minor += 1
    return Version('%d.%d.0' % (major, minor))



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for OnStateTransition class."""

from typing import Callable
from typing import Optional
from typing import Text

from launch.event import Event
from launch.event_handler import EventHandler
from launch.some_actions_type import SomeActionsType
from launch.some_substitutions_type import SomeSubstitutionsType

from ..actions import LifecycleNode
from ..events.lifecycle import StateTransition


class OnStateTransition(EventHandler):
    """Convenience class for handling a state transition of a lifecycle node."""

    def __init__(
        self,
        *,
        entities: SomeActionsType,
        target_lifecycle_node: LifecycleNode = None,
        transition: Optional[SomeSubstitutionsType] = None,
        start_state: Optional[SomeSubstitutionsType] = None,
        goal_state: Optional[SomeSubstitutionsType] = None,
        matcher: Optional[Callable[[Event], bool]] = None,
        **kwargs
    ) -> None:
        """
        Create an OnStateTransition event handler.

        There are several matching options, each of which is compared with the
        event and must match it to have the handler handle the event.
        Passing None for any of them will prevent that matching option from
        being considered (and therefore not required) when matching the event.

        If matcher is given, the other conditions are not considered.
        """
        if not isinstance(target_lifecycle_node, (LifecycleNode, type(None))):
            raise RuntimeError("OnStateTransition requires a 'LifecycleNode' action as the target")
        # Handle optional matcher argument.
        self.__custom_matcher = matcher
        if self.__custom_matcher is None:
            self.__custom_matcher = (
                lambda event: (
                    isinstance(event, StateTransition) and (
                        target_lifecycle_node is None or
                        event.action == target_lifecycle_node
                    ) and (
                        transition is None or
                        event.transition == transition
                    ) and (
                        start_state is None or
                        event.start_state == start_state
                    ) and (
                        goal_state is None or
                        event.goal_state == goal_state
                    )
                )
            )
        # Call parent init.
        super().__init__(
            matcher=self.__custom_matcher,
            entities=entities,
            **kwargs
        )
        self.__target_lifecycle_node = target_lifecycle_node

    @property
    def handler_description(self) -> Text:
        """Return the string description of the handler."""
        return '<actions>'

    @property
    def matcher_description(self):
        """Return the string description of the matcher."""
        # TODO(dhood): improve this to print more matcher properties
        if self.__target_lifecycle_node is None:
            return 'event == StateTransition'
        return 'event == StateTransition and event.action == LifecycleNode({})'.format(
            hex(id(self.__target_lifecycle_node))
        )



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with utility for evaluating parameters in a launch context."""


from collections.abc import Mapping
from collections.abc import Sequence
import pathlib
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from launch.launch_context import LaunchContext
from launch.substitution import Substitution
from launch.utilities import ensure_argument_type
from launch.utilities import perform_substitutions

import yaml

from ..parameter_descriptions import Parameter as ParameterDescription
from ..parameter_descriptions import ParameterFile
from ..parameter_descriptions import ParameterValue as ParameterValueDescription
from ..parameters_type import EvaluatedParameters
from ..parameters_type import EvaluatedParameterValue
from ..parameters_type import Parameters
from ..parameters_type import ParametersDict


def evaluate_parameter_dict(
    context: LaunchContext,
    parameters: ParametersDict
) -> Dict[str, EvaluatedParameterValue]:
    def check_sequence_type_is_allowed(sequence):
        # Check if the items of the sequence aren't dissimilar.
        # Also, check if the item type is one of bool, int, float, str.
        if not sequence:  # Don't allow empty sequences
            return False
        subtype = type(sequence[0])
        if subtype not in (bool, int, float, str):
            return False
        for item in sequence:
            if not isinstance(item, subtype):
                return False
        return True

    if not isinstance(parameters, Mapping):
        raise TypeError('expected dict')
    output_dict: Dict[str, EvaluatedParameterValue] = {}
    for name, value in parameters.items():
        if not isinstance(name, tuple):
            raise TypeError('Expecting tuple of substitutions got {}'.format(repr(name)))
        evaluated_name: str = perform_substitutions(context, list(name))
        evaluated_value: Optional[EvaluatedParameterValue] = None

        if isinstance(value, tuple) and len(value):
            if isinstance(value[0], Substitution):
                # Value is a list of substitutions, so perform them to make a string
                evaluated_value = perform_substitutions(context, list(value))

                # Handle special case where yaml.safe_load will return None given an empty string
                if len(evaluated_value) == 0:
                    evaluated_value = "''"

                try:
                    yaml_evaluated_value = yaml.safe_load(evaluated_value)
                except yaml.YAMLError:
                    raise TypeError(
                        'Unable to parse the value of parameter {} as yaml. '
                        'If the parameter is meant to be a string, try wrapping it in '
                        'launch_ros.parameter_descriptions.ParameterValue'
                        '(value, value_type=str)'.format(evaluated_name)
                    )

                if type(yaml_evaluated_value) in (bool, int, float, str, bytes):
                    evaluated_value = yaml_evaluated_value
                elif isinstance(yaml_evaluated_value, Sequence):
                    # str and bytes were already handled in the previous case
                    # If it is a list with dissimilar types, don't evaluate the value as yaml.
                    if not check_sequence_type_is_allowed(yaml_evaluated_value):
                        raise TypeError(
                            'Expected a non-empty sequence, with items of uniform type. '
                            'Allowed sequence item types are bool, int, float, str.'
                        )
                    evaluated_value = tuple(yaml_evaluated_value)
                else:
                    raise TypeError(
                        'Allowed value types are bytes, bool, int, float, str, Sequence[bool]'
                        ', Sequence[int], Sequence[float], Sequence[str]. Got {}.'
                        'If the parameter is meant to be a string, try wrapping it in '
                        'launch_ros.parameter_descriptions.ParameterValue'
                        '(value, value_type=str)'.format(type(yaml_evaluated_value))
                    )
            elif isinstance(value[0], Sequence):
                # Value is an array of a list of substitutions
                output_subvalue: List[str] = []
                for subvalue in value:
                    value = perform_substitutions(context, list(subvalue))
                    output_subvalue.append(value)
                evaluated_value = tuple(output_subvalue)
                # All values in a list must have the same type.
                # If they don't then assume it is a list of strings
                yaml_evaluated_value = [
                    yaml.safe_load(item) for item in evaluated_value
                ]
                if not check_sequence_type_is_allowed(yaml_evaluated_value):
                    raise TypeError(
                        'Expected a non-empty sequence, with items of uniform type. '
                        'Allowed sequence item types are bool, int, float, str.'
                    )
                evaluated_value = tuple(yaml_evaluated_value)
            else:
                # Value is an array of the same type, so nothing to evaluate.
                output_value = []
                target_type = type(value[0])
                for i, subvalue in enumerate(value):
                    output_value.append(target_type(subvalue))
                    evaluated_value = tuple(output_value)
        elif isinstance(value, ParameterValueDescription):
            evaluated_value = value.evaluate(context)
        else:
            # Value is a singular type, so nothing to evaluate
            ensure_argument_type(value, (float, int, str, bool, bytes), 'value')
            evaluated_value = cast(Union[float, int, str, bool, bytes], value)
        if evaluated_value is None:
            raise TypeError('given unnormalized parameters %r, %r' % (name, value))
        output_dict[evaluated_name] = evaluated_value
    return output_dict


def evaluate_parameters(context: LaunchContext, parameters: Parameters) -> EvaluatedParameters:
    """
    Evaluate substitutions to produce paths and name/value pairs.

    The parameters must have been normalized with normalize_parameters() prior to calling this.
    Substitutions for parameter values in dictionaries will be evaluated according to yaml rules.
    If you want the substitution to stay a string, the output of the substition must have quotes.

    :param parameters: normalized parameters
    :returns: values after evaluating lists of substitutions
    """
    output_params = []  # type: List[Union[pathlib.Path, Dict[str, EvaluatedParameterValue]]]
    for param in parameters:
        if isinstance(param, ParameterFile):
            # Evaluate a list of Substitution to a file path
            output_params.append(param.evaluate(context))
        elif isinstance(param, ParameterDescription):
            output_params.append(param)
        elif isinstance(param, Mapping):
            # It's a list of name/value pairs
            output_params.append(evaluate_parameter_dict(context, param))
    return tuple(output_params)



# Copyright 2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for a description of a Parameter."""

import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List
from typing import Optional
from typing import Text
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

from launch import LaunchContext
from launch import SomeSubstitutionsType
from launch import SomeSubstitutionsType_types_tuple
from launch.frontend.parse_substitution import parse_substitution
from launch.substitution import Substitution
from launch.substitutions import SubstitutionFailure
from launch.utilities import ensure_argument_type
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions
from launch.utilities.type_utils import AllowedTypesType
from launch.utilities.type_utils import normalize_typed_substitution
from launch.utilities.type_utils import perform_typed_substitution
from launch.utilities.type_utils import SomeValueType
from launch.utilities.typing_file_path import FilePath

import yaml

if TYPE_CHECKING:
    from .parameters_type import EvaluatedParameterValue


class ParameterValue:
    """Describes a ROS parameter value."""

    def __init__(
        self,
        value: SomeValueType,
        *,
        value_type: Optional[AllowedTypesType] = None
    ) -> None:
        """
        Construct a parameter value description.

        :param value: Value or substitution that can be resolved to a value.
        :param value_type: Used when `value` is a substitution, to coerce the result.
            Can be one of:
                - Scalar types: int, bool, float, str
                - List types: List[int], List[bool], List[float], List[str]
                - None: will use yaml rules, and check that the result matches one of the above.
        """
        self.__value = normalize_typed_substitution(value, value_type)
        self.__value_type = value_type
        self.__evaluated_parameter_value: Optional['EvaluatedParameterValue'] = None

    @property
    def value(self) -> SomeValueType:
        """Getter for parameter value."""
        if self.__evaluated_parameter_value is not None:
            return self.__evaluated_parameter_value
        return self.__value

    @property
    def value_type(self) -> AllowedTypesType:
        """Getter for parameter value type."""
        return self.__value_type

    def __str__(self) -> Text:
        return (
            'launch_ros.description.ParameterValue'
            f'(value={self.value}, value_type={self.value_type})'
        )

    def evaluate(self, context: LaunchContext) -> 'EvaluatedParameterValue':
        """Evaluate and return parameter rule."""
        self.__evaluated_parameter_value = perform_typed_substitution(
            context, self.value, self.value_type)
        return self.__evaluated_parameter_value


class Parameter:
    """Describes a ROS parameter."""

    def __init__(
        self,
        name: SomeSubstitutionsType,
        value: SomeValueType,
        *,
        value_type: Optional[AllowedTypesType] = None
    ) -> None:
        """
        Construct a parameter description.

        :param name: Name of the parameter.
        :param value: Value of the parameter.
        :param value_type: Used when `value` is a substitution, to coerce the result.
            Can be one of:
                - A scalar type: `int`, `str`, `float`, `bool`.
                  `bool` are written like in `yaml`.
                  Both `1` and `1.` are valid floats.
                - An uniform list: `List[int]`, `List[str]`, `List[float]`, `List[bool]`.
                  Lists are written like in `yaml`.
                - `None`, which means that yaml rules will be used.
                  The result of the convertion must be one of the above types,
                  if not `ValueError` is raised.
            If value is not a substitution and this parameter is provided,
            it will be used to check `value` type.
        """
        ensure_argument_type(name, SomeSubstitutionsType_types_tuple, 'name')

        self.__name = normalize_to_list_of_substitutions(name)
        self.__parameter_value = ParameterValue(value, value_type=value_type)
        self.__evaluated_parameter_name: Optional[Text] = None
        self.__evaluated_parameter_rule: Optional[Tuple[Text, 'EvaluatedParameterValue']] = None

    @property
    def name(self) -> Union[List[Substitution], Text]:
        """Getter for parameter name."""
        if self.__evaluated_parameter_name is not None:
            return self.__evaluated_parameter_name
        return self.__name

    @property
    def value(self) -> SomeValueType:
        """Getter for parameter value."""
        return self.__parameter_value.value

    @property
    def value_type(self) -> AllowedTypesType:
        """Getter for parameter value type."""
        return self.__parameter_value.value_type

    def __str__(self) -> Text:
        return (
            'launch_ros.description.Parameter'
            f'(name={self.name}, value={self.value}, value_type={self.value_type})'
        )

    def evaluate(self, context: LaunchContext) -> Tuple[Text, 'EvaluatedParameterValue']:
        """Evaluate and return parameter rule."""
        if self.__evaluated_parameter_rule is not None:
            return self.__evaluated_parameter_rule

        name = perform_substitutions(context, self.name)
        value = self.__parameter_value.evaluate(context)

        self.__evaluated_parameter_name = name
        self.__evaluated_parameter_rule = (name, value)
        return (name, value)


class ParameterFile:
    """Describes a ROS parameter file."""

    def __init__(
        self,
        param_file: Union[FilePath, SomeSubstitutionsType],
        *,
        allow_substs: [bool, SomeSubstitutionsType] = False
    ) -> None:
        """
        Construct a parameter file description.

        :param param_file: Path to a parameter file.
        :param allow_subst: Allow substitutions in the parameter file.
        """
        # In Python, __del__ is called even if the constructor throws an
        # exception.  It is possible for ensure_argument_type() below to
        # throw an exception and try to access these member variables
        # during cleanup, so make sure to initialize them here.
        self.__evaluated_param_file: Optional[Path] = None
        self.__created_tmp_file = False

        ensure_argument_type(
            param_file,
            SomeSubstitutionsType_types_tuple + (os.PathLike, bytes),
            'param_file',
            'ParameterFile()'
        )
        ensure_argument_type(
            allow_substs,
            bool,
            'allow_subst',
            'ParameterFile()'
        )
        self.__param_file: Union[List[Substitution], FilePath] = param_file
        if isinstance(param_file, SomeSubstitutionsType_types_tuple):
            self.__param_file = normalize_to_list_of_substitutions(param_file)
        self.__allow_substs = normalize_typed_substitution(allow_substs, data_type=bool)
        self.__evaluated_allow_substs: Optional[bool] = None

    @property
    def param_file(self) -> Union[FilePath, List[Substitution]]:
        """Getter for parameter file."""
        if self.__evaluated_param_file is not None:
            return self.__evaluated_param_file
        return self.__param_file

    @property
    def allow_substs(self) -> Union[bool, List[Substitution]]:
        """Getter for allow substitutions argument."""
        if self.__evaluated_allow_substs is not None:
            return self.__evaluated_allow_substs
        return self.__allow_substs

    def __str__(self) -> Text:
        return (
            'launch_ros.description.ParameterFile'
            f'(param_file={self.param_file}, allow_substs={self.allow_substs})'
        )

    def evaluate(self, context: LaunchContext) -> Path:
        """Evaluate and return a parameter file path."""
        if self.__evaluated_param_file is not None:
            return self.__evaluated_param_file

        param_file = self.__param_file
        if isinstance(param_file, list):
            # list of substitutions
            param_file = perform_substitutions(context, self.__param_file)

        allow_substs = perform_typed_substitution(context, self.__allow_substs, data_type=bool)
        param_file_path: Path = Path(param_file)
        if allow_substs:
            with open(param_file_path, 'r') as f, NamedTemporaryFile(
                mode='w', prefix='launch_params_', delete=False
            ) as h:
                parsed = perform_substitutions(context, parse_substitution(f.read()))
                try:
                    yaml.safe_load(parsed)
                except Exception:
                    raise SubstitutionFailure(
                        'The substituted parameter file is not a valid yaml file')
                h.write(parsed)
                param_file_path = Path(h.name)
                self.__created_tmp_file = True
        self.__evaluated_param_file = param_file_path
        return param_file_path

    def cleanup(self):
        """Delete created temporary files."""
        if self.__created_tmp_file and self.__evaluated_param_file is not None:
            try:
                os.unlink(self.__evaluated_param_file)
            except FileNotFoundError:
                pass
            self.__evaluated_param_file = None

    def __del__(self):
        self.cleanup()



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the `PushRosNamespace` action."""

from typing import List

from launch import Action
from launch import Substitution
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitutions import SubstitutionFailure
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions

from launch_ros.utilities import make_namespace_absolute
from launch_ros.utilities import prefix_namespace

from rclpy.validate_namespace import validate_namespace


@expose_action('push_ros_namespace')
@expose_action('push-ros-namespace')
class PushRosNamespace(Action):
    """
    Action that pushes the ros namespace.

    It's automatically popped when used inside a scoped `GroupAction`.
    There's no other way of popping it.
    """

    def __init__(
        self,
        namespace: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """Create a PushRosNamespace action."""
        super().__init__(**kwargs)
        self.__namespace = normalize_to_list_of_substitutions(namespace)

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `PushRosNamespace` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['namespace'] = parser.parse_substitution(entity.get_attr('namespace'))
        return cls, kwargs

    @property
    def namespace(self) -> List[Substitution]:
        """Getter for self.__namespace."""
        return self.__namespace

    def execute(self, context: LaunchContext):
        """Execute the action."""
        pushed_namespace = perform_substitutions(context, self.namespace)
        previous_namespace = context.launch_configurations.get('ros_namespace', None)
        namespace = make_namespace_absolute(
            prefix_namespace(previous_namespace, pushed_namespace))
        try:
            validate_namespace(namespace)
        except Exception:
            raise SubstitutionFailure(
                'The resulting namespace is invalid:'
                " [previous_namespace='{}', pushed_namespace='{}']".format(
                    previous_namespace, pushed_namespace
                )
            )
        context.launch_configurations['ros_namespace'] = namespace



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matchers for node names."""

from typing import Callable
from typing import Text

if False:
    # imports here would cause loops, but are only used as forward-references for type-checking
    from ..actions import Node  # noqa: F401


def matches_node_name(node_name: Text) -> Callable[['Node'], bool]:
    """Return a matcher which matches based on the name of the node itself."""
    if not node_name.startswith('/'):
        node_name = f'/{node_name}'
    return lambda action: action.node_name == node_name



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to help detect non-unique node names."""

from collections import defaultdict

from launch.launch_context import LaunchContext


def add_node_name(context: LaunchContext, node_name: str) -> None:
    """
    Add a node name to the context, indicating an occurrence of the node name.

    :param context: the context that keeps track of the node names
    :param node_name: the node name to keep track
    """
    try:
        unique_node_names = context.locals.unique_ros_node_names
    except AttributeError:
        context.extend_globals({'unique_ros_node_names': defaultdict(int)})
        unique_node_names = context.locals.unique_ros_node_names
    unique_node_names[node_name] += 1


def get_node_name_count(context: LaunchContext, node_name: str) -> int:
    """
    Get the number of times the node name has occurred, according to the context.

    :param context: the context that keeps track of the node names
    :param node_name: the node name to keep track
    :returns: number of times the node name has occurred
    """
    try:
        unique_node_names = context.locals.unique_ros_node_names
    except AttributeError:
        return 0
    return unique_node_names[node_name]



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""actions Module."""

from .composable_node_container import ComposableNodeContainer
from .lifecycle_node import LifecycleNode
from .load_composable_nodes import LoadComposableNodes
from .node import Node
from .push_ros_namespace import PushRosNamespace
from .ros_timer import RosTimer
from .set_parameter import SetParameter
from .set_parameters_from_file import SetParametersFromFile
from .set_remap import SetRemap
from .set_use_sim_time import SetUseSimTime


__all__ = [
    'ComposableNodeContainer',
    'LifecycleNode',
    'LoadComposableNodes',
    'Node',
    'PushRosNamespace',
    'RosTimer',
    'SetParameter',
    'SetParametersFromFile',
    'SetRemap',
    'SetUseSimTime'
]



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for ROS parameter types."""

import pathlib

from typing import Any
from typing import Dict
from typing import Mapping
from typing import Sequence
from typing import Union

from launch.some_substitutions_type import SomeSubstitutionsType
from launch.some_substitutions_type import SomeSubstitutionsType_types_tuple
from launch.substitution import Substitution

from .parameter_descriptions import Parameter as ParameterDescription
from .parameter_descriptions import ParameterFile
from .parameter_descriptions import ParameterValue as ParameterValueDescription


# Parameter value types used below
_SingleValueType_types_tuple = (str, int, float, bool)
_SingleValueType = Union[str, int, float, bool]
_MultiValueType = Union[
    Sequence[str], Sequence[int], Sequence[float], Sequence[bool], bytes]

SomeParameterFile = Union[SomeSubstitutionsType, pathlib.Path, ParameterFile]
SomeParameterName = Sequence[Union[Substitution, str]]
SomeParameterValue = Union[
    ParameterValueDescription,
    SomeSubstitutionsType,
    Sequence[SomeSubstitutionsType],
    _SingleValueType,
    _MultiValueType
]
SomeParameterValue_types_tuple = (
    SomeSubstitutionsType_types_tuple +
    _SingleValueType_types_tuple +
    (bytes,)
)

# TODO(sloretz) Recursive type when mypy supports them python/mypy#731
_SomeParametersDict = Mapping[SomeParameterName, Any]
SomeParametersDict = Mapping[SomeParameterName, Union[SomeParameterValue, _SomeParametersDict]]

# Paths to yaml files with parameters, or dictionaries of parameters, or pairs of
# parameter names and values
SomeParameters = Sequence[Union[SomeParameterFile, ParameterDescription, SomeParametersDict]]

ParameterFile = ParameterFile  # re-export
ParameterName = Sequence[Substitution]
ParameterValue = Union[
    Sequence[Substitution],
    Sequence[Sequence[Substitution]],
    _SingleValueType,
    _MultiValueType,
    ParameterValueDescription]

# Normalized (flattened to avoid having a recursive type) parameter dict
ParametersDict = Dict[ParameterName, ParameterValue]

# Normalized parameters
Parameters = Sequence[Union[ParameterFile, ParametersDict, ParameterDescription]]

EvaluatedParameterValue = Union[_SingleValueType, _MultiValueType]
# Evaluated parameters: filenames or dictionary after substitutions have been evaluated
EvaluatedParameters = Sequence[
    Union[pathlib.Path, ParameterDescription, Dict[str, EvaluatedParameterValue]]
]



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from launch import Action
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext

from launch_ros.ros_adapters import get_ros_node

from rclpy.parameter import Parameter


@expose_action('set_use_sim_time')
class SetUseSimTime(Action):
    """Action that sets the 'use_sim_time' parameter in the current context."""

    def __init__(
        self,
        value: bool,
        **kwargs
    ) -> None:
        """Create a SetUseSimTime action."""
        super().__init__(**kwargs)
        self.__value = value

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `SetUseSimTime` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['value'] = parser.parse_substitution(entity.get_attr('value'))
        return cls, kwargs

    @property
    def value(self) -> bool:
        """Getter for value."""
        return self.__value

    def execute(self, context: LaunchContext):
        """Execute the action."""
        node = get_ros_node(context)
        param = Parameter(
            'use_sim_time',
            Parameter.Type.BOOL,
            self.value
        )
        node.set_parameters([param])
        if not node.get_parameter('use_sim_time').get_parameter_value().bool_value:
            raise RuntimeError('Failed to set use_sim_time parameter')



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for a description of a ComposableNode."""

from typing import List
from typing import Optional

from launch.frontend import Entity
from launch.frontend import Parser
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitution import Substitution
# from launch.utilities import ensure_argument_type
from launch.utilities import normalize_to_list_of_substitutions
from launch_ros.parameters_type import Parameters
from launch_ros.parameters_type import SomeParameters
from launch_ros.remap_rule_type import RemapRules
from launch_ros.remap_rule_type import SomeRemapRules
from launch_ros.utilities import normalize_parameters
from launch_ros.utilities import normalize_remap_rules


class ComposableNode:
    """Describes a ROS node that can be loaded into a container with other nodes."""

    def __init__(
        self, *,
        package: SomeSubstitutionsType,
        plugin: SomeSubstitutionsType,
        name: Optional[SomeSubstitutionsType] = None,
        namespace: Optional[SomeSubstitutionsType] = None,
        parameters: Optional[SomeParameters] = None,
        remappings: Optional[SomeRemapRules] = None,
        extra_arguments: Optional[SomeParameters] = None,
    ) -> None:
        """
        Initialize a ComposableNode description.

        :param package: name of the ROS package the node plugin lives in
        :param plugin: name of the plugin to be loaded
        :param name: name to give to the ROS node
        :param namespace: namespace to give to the ROS node
        :param parameters: list of either paths to yaml files or dictionaries of parameters
        :param remappings: list of from/to pairs for remapping names
        :param extra_arguments: container specific arguments to be passed to the loaded node
        """
        self.__package = normalize_to_list_of_substitutions(package)
        self.__node_plugin = normalize_to_list_of_substitutions(plugin)

        self.__node_name = None  # type: Optional[List[Substitution]]
        if name is not None:
            self.__node_name = normalize_to_list_of_substitutions(name)

        self.__node_namespace = None  # type: Optional[List[Substitution]]
        if namespace is not None:
            self.__node_namespace = normalize_to_list_of_substitutions(namespace)

        self.__parameters = None  # type: Optional[Parameters]
        if parameters is not None:
            self.__parameters = normalize_parameters(parameters)

        self.__remappings = None  # type: Optional[RemapRules]
        if remappings:
            self.__remappings = normalize_remap_rules(remappings)

        self.__extra_arguments = None  # type: Optional[Parameters]
        if extra_arguments:
            self.__extra_arguments = normalize_parameters(extra_arguments)

    @classmethod
    def parse(cls, parser: Parser, entity: Entity):
        """Parse composable_node."""
        from launch_ros.actions import Node
        kwargs = {}

        kwargs['package'] = parser.parse_substitution(entity.get_attr('pkg'))
        kwargs['plugin'] = parser.parse_substitution(entity.get_attr('plugin'))
        kwargs['name'] = parser.parse_substitution(entity.get_attr('name'))

        namespace = entity.get_attr('namespace', optional=True)
        if namespace is not None:
            kwargs['namespace'] = parser.parse_substitution(namespace)

        parameters = entity.get_attr('param', data_type=List[Entity], optional=True)
        if parameters is not None:
            kwargs['parameters'] = Node.parse_nested_parameters(parameters, parser)

        remappings = entity.get_attr('remap', data_type=List[Entity], optional=True)
        if remappings is not None:
            kwargs['remappings'] = [
                (
                    parser.parse_substitution(remap.get_attr('from')),
                    parser.parse_substitution(remap.get_attr('to'))
                ) for remap in remappings
            ]

            for remap in remappings:
                remap.assert_entity_completely_parsed()

        extra_arguments = entity.get_attr('extra_arg', data_type=List[Entity], optional=True)
        if extra_arguments is not None:
            kwargs['extra_arguments'] = [
                {
                    tuple(parser.parse_substitution(extra_arg.get_attr('name'))):
                    parser.parse_substitution(extra_arg.get_attr('value'))
                } for extra_arg in extra_arguments
            ]

            for extra_arg in extra_arguments:
                extra_arg.assert_entity_completely_parsed()

        entity.assert_entity_completely_parsed()

        return cls, kwargs

    @property
    def package(self) -> List[Substitution]:
        """Get node package name as a sequence of substitutions to be performed."""
        return self.__package

    @property
    def node_plugin(self) -> List[Substitution]:
        """Get node plugin name as a sequence of substitutions to be performed."""
        return self.__node_plugin

    @property
    def node_name(self) -> Optional[List[Substitution]]:
        """Get node name as a sequence of substitutions to be performed."""
        return self.__node_name

    @property
    def node_namespace(self) -> Optional[List[Substitution]]:
        """Get node namespace as a sequence of substitutions to be performed."""
        return self.__node_namespace

    @property
    def parameters(self) -> Optional[Parameters]:
        """Get node parameter YAML files or dicts with substitutions to be performed."""
        return self.__parameters

    @property
    def remappings(self) -> Optional[RemapRules]:
        """Get node remapping rules as (from, to) tuples with substitutions to be performed."""
        return self.__remappings

    @property
    def extra_arguments(self) -> Optional[Parameters]:
        """Get container extra arguments YAML files or dicts with substitutions to be performed."""
        return self.__extra_arguments



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the FindPackage substitution."""

from typing import Iterable
from typing import List
from typing import Text

from ament_index_python.packages import get_package_prefix
from ament_index_python.packages import get_package_share_directory

from launch.frontend import expose_substitution
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitution import Substitution
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions


class FindPackage(Substitution):
    """
    Abstract base class for substitutions involving finding a package.

    Subclasses should implement the find() method.
    """

    def __init__(
        self,
        package: SomeSubstitutionsType,
    ) -> None:
        """Create a FindPackage substitution."""
        super().__init__()
        self.__package = normalize_to_list_of_substitutions(package)

    @classmethod
    def parse(cls, data: Iterable[SomeSubstitutionsType]):
        """Parse a FindPackage substitution."""
        if not data or len(data) != 1:
            raise AttributeError('find package substitutions expect 1 argument')
        kwargs = {'package': data[0]}
        return cls, kwargs

    @property
    def package(self) -> List[Substitution]:
        """Getter for package."""
        return self.__package

    def find(self, package_name: Text) -> Text:
        """
        Find a directory for a package.

        Called when the substitution is performed.

        :param: package_name The name of the package.
        :return: A directory related to the package.
        """
        raise NotImplementedError()

    def describe(self) -> Text:
        """Return a description of this substitution as a string."""
        pkg_str = ' + '.join([sub.describe() for sub in self.package])
        return '{}(pkg={})'.format(self.__class__.__name__, pkg_str)

    def perform(self, context: LaunchContext) -> Text:
        """Perform the substitution by locating the package."""
        package = perform_substitutions(context, self.package)
        result = self.find(package)
        return result


@expose_substitution('find-pkg-prefix')
class FindPackagePrefix(FindPackage):
    """
    Substitution that tries to locate the package prefix of a ROS package.

    The ROS package is located using ament_index_python.

    :raise: ament_index_python.packages.PackageNotFoundError when package is
        not found during substitution.
    """

    def find(self, package_name: Text) -> Text:
        """Find the package prefix."""
        return get_package_prefix(package_name)


@expose_substitution('find-pkg-share')
class FindPackageShare(FindPackage):
    """
    Substitution that tries to locate the share directory of a ROS package.

    The directory is located using ament_index_python.

    :raise: ament_index_python.packages.PackageNotFoundError when package is
        not found during substitution.
    """

    def find(self, package_name: Text) -> Text:
        """Find the share directory of a package."""
        return get_package_share_directory(package_name)



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for standard "lifecycle_node_matchers", which are used with targeted lifecycle events."""

from typing import Callable
from typing import Text
import warnings

from ..matchers import matches_node_name as _matches_node_name

if False:
    # imports here would cause loops, but are only used as forward-references for type-checking
    from ...actions import LifecycleNode  # noqa: F401


def matches_node_name(node_name: Text) -> Callable[['LifecycleNode'], bool]:
    """Return a matcher which matches based on the name of the node itself."""
    warnings.warn(
        "'matches_node_name' has been moved into the 'launch.events' module and will be removed "
        "from the 'lifecycle' module in the future"
    )
    return _matches_node_name(node_name)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with utility to transform evaluated parameters into parameter lists."""

import pathlib
from typing import List
import warnings

from launch.launch_context import LaunchContext
import rclpy.parameter

import yaml

from .evaluate_parameters import evaluate_parameter_dict
from .normalize_parameters import normalize_parameter_dict

from ..parameters_type import EvaluatedParameters


def __normalize_parameters_dict(dictionary):
    """
    Combine namespaces and their node names into one key and remove the "ros__parameters" key.

    Return an empty dict if the "ros__parameters" key is not found.
    """
    # Recursive function
    def normalize_parameters_dict(dictionary, keys, result_dict):
        for key, value in dictionary.items():
            # Base case: found 'ros__parameters' key
            # Return
            if key == 'ros__parameters':
                node_name = '/'.join(keys)
                result_dict[node_name] = value
                return result_dict

            if isinstance(value, dict):
                keys.append(key.lstrip('/'))
                result_dict = normalize_parameters_dict(value, keys, result_dict)
                # Reset keys in case there are multiple ros__parameter entries
                keys = []

        return result_dict

    return normalize_parameters_dict(dictionary, [], {})


def to_parameters_list(
    context: LaunchContext,
    node_name: str,
    namespace: str,
    evaluated_parameters: EvaluatedParameters
) -> List[rclpy.parameter.Parameter]:
    """
    Transform evaluated parameters into a list of rclpy.parameter.Parameter objects.

    :param context: to carry out substitutions during normalization of parameter files.
        See `normalize_parameters()` documentation for further reference.
    :param node_name: node name
    :param namespace: node namespace
    :param parameters: parameters as either paths to parameter files or name/value pairs.
        See `evaluate_parameters()` documentation for further reference.
    :returns: a list of parameters
    """
    parameters = []  # type: List[rclpy.parameter.Parameter]
    node_name = node_name.lstrip('/')
    if namespace and namespace != '/':
        namespace = namespace.strip('/')
        node_name = f'{namespace}/{node_name}'

    params_set = {}
    warned_once = False
    for params_set_or_path in evaluated_parameters:
        if isinstance(params_set_or_path, pathlib.Path):
            with open(str(params_set_or_path), 'r') as f:
                param_dict = yaml.safe_load(f)
                normalized_param_dict = __normalize_parameters_dict(param_dict)

                if normalized_param_dict:
                    param_dict.clear()
                    if '**' in normalized_param_dict:
                        param_dict = normalized_param_dict['**']
                    if node_name in normalized_param_dict:
                        param_dict.update(normalized_param_dict[node_name])
                    if not warned_once and not node_name:
                        warnings.warn(
                            'node name not provided to launch; parameter files will not apply',
                            UserWarning
                        )
                        warned_once = True

                if param_dict:
                    params_set = evaluate_parameter_dict(
                        context, normalize_parameter_dict(param_dict)
                    )
        else:
            params_set = params_set_or_path
        if not isinstance(params_set, dict):
            raise RuntimeError('invalid evaluated parameters {}'.format(repr(params_set)))
        for name, value in params_set.items():
            type_ = next((
                type_ for type_ in rclpy.parameter.Parameter.Type if type_.check(value)
            ), None)  # type: rclpy.parameter.Parameter.Type
            if type_ is None:
                raise RuntimeError('invalid parameter value {}'.format(repr(value)))
            parameters.append(rclpy.parameter.Parameter(name, type_, value))
    return parameters



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the ComposableNodeContainer action."""

from typing import List
from typing import Optional

from launch.action import Action
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType

from .node import Node

from ..descriptions import ComposableNode


@expose_action('node_container')
class ComposableNodeContainer(Node):
    """Action that executes a container ROS node for composable ROS nodes."""

    def __init__(
        self,
        *,
        name: SomeSubstitutionsType,
        namespace: SomeSubstitutionsType,
        composable_node_descriptions: Optional[List[ComposableNode]] = None,
        **kwargs
    ) -> None:
        """
        Construct a ComposableNodeContainer action.

        Most arguments are forwarded to :class:`launch_ros.actions.Node`, so see the documentation
        of that class for further details.

        :param: name the name of the node, mandatory for full container node name resolution
        :param: namespace the ROS namespace for this Node, mandatory for full container node
             name resolution
        :param composable_node_descriptions: optional descriptions of composable nodes to be loaded
        """
        super().__init__(name=name, namespace=namespace, **kwargs)
        self.__composable_node_descriptions = composable_node_descriptions

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Parse node_container."""
        _, kwargs = super().parse(entity, parser)

        composable_nodes = entity.get_attr(
            'composable_node', data_type=List[Entity], optional=True)
        if composable_nodes is not None:
            kwargs['composable_node_descriptions'] = []
            for entity in composable_nodes:
                composable_node_cls, composable_node_kwargs = ComposableNode.parse(parser, entity)
                kwargs['composable_node_descriptions'].append(
                    composable_node_cls(**composable_node_kwargs))

        return cls, kwargs

    def execute(self, context: LaunchContext) -> Optional[List[Action]]:
        """
        Execute the action.

        Most work is delegated to :meth:`launch_ros.actions.Node.execute`, except for the
        composable nodes load action if it applies.
        """
        load_actions = None  # type: Optional[List[Action]]
        if (
            self.__composable_node_descriptions is not None and
            len(self.__composable_node_descriptions) > 0
        ):
            from .load_composable_nodes import LoadComposableNodes
            # Perform load action once the container has started.
            load_actions = [
                LoadComposableNodes(
                    composable_node_descriptions=self.__composable_node_descriptions,
                    target_container=self
                )
            ]
        container_actions = super().execute(context)  # type: Optional[List[Action]]
        if container_actions is not None and load_actions is not None:
            return container_actions + load_actions
        if container_actions is not None:
            return container_actions
        if load_actions is not None:
            return load_actions
        return None
